"""
tests for HDF dataset
"""

from __future__ import annotations

from typing import Dict, Any, List, Tuple, Optional
import os
import sys
import _setup_test_env  # noqa

from returnn.datasets import Dataset
from returnn.datasets.hdf import *
import returnn.util.basic as util
import h5py
import numpy as np
import os
import unittest
from returnn.util import better_exchook


my_dir = os.path.dirname(os.path.abspath(__file__))


def test_HDFDataset_init():
    """
    This method tests initialization of the HDFDataset class
    """
    toy_dataset = HDFDataset()
    assert toy_dataset.file_start == [0], "self.file_start init problem, should be [0]"
    assert toy_dataset.files == [], "self.files init problem, should be []"
    assert toy_dataset.file_seq_start == [], "self.file_seq_start init problem, should be []"
    return toy_dataset


def generate_dummy_hdf(num_datasets=1):
    filenames = []
    for idx in range(1, num_datasets + 1):
        fn = get_test_tmp_file(".%i.hdf5" % idx)
        filenames.append(fn)
        dataset = h5py.File(fn, "w")
        dataset.create_group("streams")

        dataset["streams"].create_group("features")
        dataset["streams"]["features"].attrs["parser"] = "feature_sequence"
        dataset["streams"]["features"].create_group("data")

        dataset["streams"].create_group("classes")
        dataset["streams"]["classes"].attrs["parser"] = "sparse"
        dataset["streams"]["classes"].create_group("data")

        import string

        random = np.random.RandomState(42)
        feature_size = 13
        seq_names = ["dataset_%d_sequence_%d" % (idx, i) for i in range(100)]
        class_names = list(string.ascii_lowercase)
        num_classes = len(class_names)
        print(class_names, len(class_names))
        for idx in range(100):
            class_id = random.randint(low=0, high=num_classes)
            seq_len = random.randint(low=1, high=20)
            features = random.rand(seq_len, feature_size)
            dataset["streams"]["features"]["data"].create_dataset(name=seq_names[idx], data=features, dtype="float32")
            dataset["streams"]["classes"]["data"].create_dataset(
                name=seq_names[idx], shape=(1,), data=np.int32(class_id), dtype="int32"
            )

        dt = h5py.special_dtype(vlen=str)
        feature_names = dataset["streams"]["classes"].create_dataset(
            "feature_names", shape=(len(class_names),), dtype=dt
        )
        for id_x, orth in enumerate(class_names):
            feature_names[id_x] = orth

        dt = h5py.special_dtype(vlen=str)
        sequence_names_data = dataset.create_dataset("seq_names", shape=(len(seq_names),), dtype=dt)
        for ind, val in enumerate(seq_names):
            sequence_names_data[ind] = val

        dataset.close()
    return filenames


# Note that nosetests might even call this function, as it has "test" in its name... Does not matter, though.
def get_test_tmp_file(suffix=".hdf"):
    """
    :param str suffix: e.g. ".hdf"
    :return: filename
    :rtype: str
    """
    import tempfile

    f = tempfile.NamedTemporaryFile(suffix=suffix, delete=False)
    f.close()
    fn = f.name
    import atexit

    atexit.register(lambda: os.remove(fn))
    return fn


_hdf_cache = {}  # opts -> hdf fn


def generate_hdf_from_other(opts: Dict[str, Any], suffix: str = ".hdf", *, use_cache: bool = True) -> str:
    """
    :param opts:
    :param suffix:
    :param use_cache:
    :return: hdf filename
    """
    # See test_hdf_dump.py and tools/hdf_dump.py.
    from returnn.util.basic import make_hashable

    cache_key = None
    if use_cache:
        cache_key = make_hashable(opts)
        if cache_key in _hdf_cache:
            return _hdf_cache[cache_key]
    fn = get_test_tmp_file(suffix=suffix)
    from returnn.datasets.basic import init_dataset

    dataset = init_dataset(opts)
    hdf_dataset = HDFDatasetWriter(fn)
    hdf_dataset.dump_from_dataset(dataset)
    hdf_dataset.close()
    if use_cache:
        _hdf_cache[cache_key] = fn
    return fn


def generate_hdf_from_dummy():
    """
    :return: hdf filename
    :rtype: str
    """
    return generate_hdf_from_other(
        {"class": "DummyDataset", "input_dim": 13, "output_dim": 7, "num_seqs": 23, "seq_len": 17}
    )


def test_hdf_dump():
    generate_hdf_from_dummy()


class DatasetTestReader:
    def __init__(self, dataset):
        """
        :param Dataset dataset:
        """
        self.dataset = dataset
        self.data_keys = None
        self.data_shape = {}  # key -> shape
        self.data_sparse = {}  # key -> bool
        self.data_dtype = {}  # key -> str
        self.data = {}  # type: typing.Dict[str,typing.List[numpy.ndarray]]  # key -> list
        self.seq_lens = []  # type: typing.List[util.NumbersDict]
        self.seq_tags = []
        self.num_seqs = 0

    def read_all(self, epoch=1):
        """
        :param int epoch:
        """
        dataset = self.dataset
        dataset.init_seq_order(epoch=epoch)
        data_keys = dataset.get_data_keys()
        self.data_keys = data_keys
        self.data_shape = {key: dataset.get_data_shape(key) for key in data_keys}
        self.data_sparse = {key: dataset.is_data_sparse(key) for key in data_keys}
        self.data_dtype = {key: dataset.get_data_dtype(key) for key in data_keys}
        seq_idx = 0
        while dataset.is_less_than_num_seqs(seq_idx):
            dataset.load_seqs(seq_idx, seq_idx + 1)
            for key in data_keys:
                data = dataset.get_data(seq_idx=seq_idx, key=key)
                self.data.setdefault(key, []).append(data)
            seq_len = dataset.get_seq_length(seq_idx=seq_idx)
            self.seq_lens.append(seq_len)
            seq_tag = dataset.get_tag(seq_idx)
            self.seq_tags.append(seq_tag)
            seq_idx += 1
        print("Iterated through %r, num seqs %i" % (dataset, seq_idx))
        self.num_seqs += seq_idx


def test_hdf_dump_not_frame_synced():
    num_seqs = 3
    from returnn.datasets.generating import TaskNumberBaseConvertDataset

    hdf_fn = generate_hdf_from_other({"class": "TaskNumberBaseConvertDataset", "num_seqs": num_seqs})
    hdf = HDFDataset([hdf_fn])
    orig = TaskNumberBaseConvertDataset(num_seqs=num_seqs)
    hdf_reader = DatasetTestReader(hdf)
    orig_reader = DatasetTestReader(orig)
    hdf_reader.read_all()
    orig_reader.read_all()
    assert hdf_reader.data_keys == orig_reader.data_keys == ["data", "classes"]
    assert hdf_reader.num_seqs == orig_reader.num_seqs == num_seqs
    for seq_idx in range(num_seqs):
        # Not synced, i.e. different lengths:
        assert orig_reader.seq_lens[seq_idx]["data"] != orig_reader.seq_lens[seq_idx]["classes"]
        for key in orig_reader.data_keys:
            assert hdf_reader.seq_lens[seq_idx][key] == orig_reader.seq_lens[seq_idx][key]
            assert hdf_reader.data[key][seq_idx].tolist() == orig_reader.data[key][seq_idx].tolist()


def test_HDFDataset_partition_epoch():
    partition_epoch = 3
    num_seqs = 11
    from returnn.datasets.generating import TaskNumberBaseConvertDataset

    hdf_fn = generate_hdf_from_other({"class": "TaskNumberBaseConvertDataset", "num_seqs": num_seqs})
    hdf = HDFDataset([hdf_fn], partition_epoch=partition_epoch)
    orig = TaskNumberBaseConvertDataset(num_seqs=num_seqs)
    hdf_reader = DatasetTestReader(hdf)
    orig_reader = DatasetTestReader(orig)
    for epoch in range(1, partition_epoch + 1):
        hdf_reader.read_all(epoch=epoch)
    orig_reader.read_all()  # single epoch
    assert hdf_reader.data_keys == orig_reader.data_keys == ["data", "classes"]
    assert hdf_reader.num_seqs == orig_reader.num_seqs == num_seqs
    for seq_idx in range(num_seqs):
        # Not synced, i.e. different lengths:
        assert orig_reader.seq_lens[seq_idx]["data"] != orig_reader.seq_lens[seq_idx]["classes"]
        for key in orig_reader.data_keys:
            assert hdf_reader.seq_lens[seq_idx][key] == orig_reader.seq_lens[seq_idx][key]
            assert hdf_reader.data[key][seq_idx].tolist() == orig_reader.data[key][seq_idx].tolist()


def test_SimpleHDFWriter():
    fn = get_test_tmp_file(suffix=".hdf")
    os.remove(fn)  # SimpleHDFWriter expects that the file does not exist
    n_dim = 13
    writer = SimpleHDFWriter(filename=fn, dim=n_dim, labels=None)
    seq_lens1 = [11, 7, 5]
    writer.insert_batch(
        inputs=numpy.random.normal(size=(len(seq_lens1), max(seq_lens1), n_dim)).astype("float32"),
        seq_len=seq_lens1,
        seq_tag=["seq-%i" % i for i in range(len(seq_lens1))],
    )
    seq_lens2 = [10, 13, 3, 2]
    writer.insert_batch(
        inputs=numpy.random.normal(size=(len(seq_lens2), max(seq_lens2), n_dim)).astype("float32"),
        seq_len=seq_lens2,
        seq_tag=["seq-%i" % (i + len(seq_lens1)) for i in range(len(seq_lens2))],
    )
    writer.close()
    seq_lens = seq_lens1 + seq_lens2

    dataset = HDFDataset(files=[fn])
    assert dataset.get_data_keys() == ["data"]
    assert dataset.get_target_list() == []
    reader = DatasetTestReader(dataset=dataset)
    reader.read_all()
    assert "data" in reader.data_keys  # "classes" might be in there as well, although not really correct/existing
    assert reader.data_sparse["data"] is False
    assert list(reader.data_shape["data"]) == [n_dim]
    assert reader.data_dtype["data"] == "float32"
    assert len(seq_lens) == reader.num_seqs
    for i, seq_len in enumerate(seq_lens):
        assert reader.seq_lens[i]["data"] == seq_len
    print("tags:", reader.seq_tags)
    assert reader.seq_tags == ["seq-%i" % i for i in range(reader.num_seqs)]
    assert isinstance(reader.seq_tags[0], str)


def test_SimpleHDFWriter_small():
    fn = get_test_tmp_file(suffix=".hdf")
    os.remove(fn)  # SimpleHDFWriter expects that the file does not exist
    n_dim = 3
    writer = SimpleHDFWriter(filename=fn, dim=n_dim, labels=None)
    seq_lens = [2, 3]
    writer.insert_batch(
        inputs=numpy.random.normal(size=(len(seq_lens), max(seq_lens), n_dim)).astype("float32"),
        seq_len=seq_lens,
        seq_tag=["seq-%i" % i for i in range(len(seq_lens))],
    )
    writer.close()

    dataset = HDFDataset(files=[fn])
    reader = DatasetTestReader(dataset=dataset)
    reader.read_all()
    assert "data" in reader.data_keys  # "classes" might be in there as well, although not really correct/existing
    assert reader.data_sparse["data"] is False
    assert list(reader.data_shape["data"]) == [n_dim]
    assert reader.data_dtype["data"] == "float32"
    assert len(seq_lens) == reader.num_seqs
    for i, seq_len in enumerate(seq_lens):
        assert reader.seq_lens[i]["data"] == seq_len

    print("raw content (gzipped):")
    import gzip

    print(repr(gzip.compress(open(fn, "rb").read())))


def test_SimpleHDFWriter_empty_extra():
    # This tests whether adding empty data in extra works
    fn = get_test_tmp_file(suffix=".hdf")
    os.remove(fn)  # SimpleHDFWriter expects that the file does not exist
    n_dim = 2
    writer = SimpleHDFWriter(filename=fn, dim=n_dim, labels=None)

    j = 0
    all_seq_lens = []
    # seed
    numpy.random.seed(42)
    while j < 1000:
        batch_size = numpy.random.randint(1, 10)
        seq_lens = numpy.random.randint(0, 8, size=batch_size)
        main_input = numpy.random.normal(size=(len(seq_lens), max(seq_lens), n_dim)).astype("float32")

        extra_input = main_input.copy() + 4.2
        assert main_input.shape == extra_input.shape

        writer.insert_batch(
            inputs=main_input,
            seq_len=seq_lens,
            seq_tag=["seq-%i" % (j + i) for i in range(len(seq_lens))],
            extra={"test-extra": extra_input},
        )
        j += len(seq_lens)
        all_seq_lens += seq_lens.tolist()

    assert 0 in all_seq_lens, "please update random seed, we expect to test empty seqs"

    writer.close()

    dataset = HDFDataset(files=[fn])
    assert dataset.get_data_keys() == ["data", "test-extra"], dataset.get_data_keys()
    assert dataset.get_target_list() == ["test-extra"]
    reader = DatasetTestReader(dataset=dataset)
    reader.read_all()
    assert "data" in reader.data_keys
    assert "test-extra" in reader.data_keys
    assert reader.data_sparse["data"] is False
    assert list(reader.data_shape["data"]) == [n_dim]
    assert reader.data_dtype["data"] == "float32"
    assert j == reader.num_seqs
    assert j == len(reader.seq_lens)

    for i, seq_len in enumerate(all_seq_lens):
        assert reader.seq_lens[i]["data"] == seq_len
        for k in range(0, seq_len):  # only test the first seq_len elements
            a = reader.data["data"][i][k] + 4.2
            b = reader.data["test-extra"][i][k]
            assert numpy.allclose(a, b), f"i={i}"

    assert reader.seq_tags == ["seq-%i" % i for i in range(reader.num_seqs)]
    assert isinstance(reader.seq_tags[0], str)


def test_read_simple_hdf():
    # n_dim, seq_lens, raw_gzipped is via test_SimpleHDFWriter_small
    n_dim = 3
    seq_lens = [2, 3]
    raw_gzipped = (
        b"\x1f\x8b\x08\x00\x80\xc8f\\\x02\xff\xed\x9a=l\xd3@\x14\xc7\xefl'XQ\x8a\xd2vhTT\xf0\x84X\x90\xc2V1\xb8TJ\xa0C"
        b'\x81\x8a\x04\xa9\x03BM\x85i"5Q\x82]$`h%\x18@'
        b"\xea\xc6\x02\x1b\x03\x03\x1f\x0b\x0b\x1bm\x10S\xa7NLL\xb01\x96\x1d"
        b"\xa9\xd8\xbewI}\x8d\x9d\xb1i\xf9\xff\x86\\\xee\xf2\x9e\xef\xc3\xff\xbb{"
        b"\xf6\xe5\xc5\\\xf1\xeaHf2\xc3\x02L\x93\x19,"
        b"\xc7\x0e\xb2O|\xb7\xa3y\xf9\xfb\x12\xa5\x9c"
        b"\xd2\xe7\x94\xbe\xd3d\xb9\x19\xfe\x96\xa7\xf2\x1c]\xdf\xd2E\xbeF\x8e"
        b"\x95[\xa5R`\xbd\xaf \xeby\x95\x12\xe9\x05\x06\xfeG\xe6J\xb3\x0bA\xbaH\xf9\x02\xa5;Z\xd4n\xb5\xba\xec\xac\xba"
        b"\x8c\xb9N{\xdei\xaex5W\x94\xd7\x9b\xad5O\x94W\xaa+nW\xaf\x83\xf44JzUu\x9de\xd3\xfe\\\t\x14{"
        b"\xda\xffn\x8a\xeb/T=\xaf\\\x7f\xec\x04:7\xfd\xe9\x14Z^\x89\xcc\x0f\x92\xbd\xefS "
        b"\x7f3\xf4o\xae5\x8a\xf5\x86\x1b\xeb\xc7\x99ZoF\xfa\xcdS"
        b"\x97\xc5\xfc\x1aX\xaf\xf4\x1f\x91\xfe\x95z\xc3q=\xa7\xe5"
        b"&\xf9\xa7\xe2\xdb]v\xda\xf1\xed\xee\xdd\x9e|\xe28s\x96\x16"
        b">\\\xe6\x85=\xe7\xbc\xaf\xbdN\xeb\xca8\x17mKQ^\xd3"
        b"\xb4\xd0\xc1$\x7f\x9d\xab+\x89`\x8cZ\x1b\x18o\xec\xdc\xbf\xd3"
        b'\xbb\xc3\xc3A\xf9\xc6\xcd"\xf7G\xda\x92:\x1dK\xb67'
        b"\xe5\xfak$\xdb\xc9\xd5\xfdg:\xd9N\xce\x8b\xd6\xd4\xf1^7\x0e\xeb"
        b"J\xf4\x8c\x0b\x99t\xf5\xa9)\xfb\x9d\xd6\xd5Y.t\r"
        b"\xf4-\x86\xd6\xa2\xf9@z\xd3YTo\x9a\xbew\x8a|\rqY\xa3\xbf\xde"
        b"ZG2\x1e\xc1>\xcb\xfb\xed\xb3V\xb2\xdf\xe6Y\xa5\xc0"
        b"\x88\x8e\x9b\x81-\n\x00\x00\x00\x00\xe0D1("
        b"\x8eN)\xcf\x99j|\xa9\xfb\xf1q`9j\x9d\xeb\xc6\xd1\x13&\x9bX\xef>_\xc6\xc6\xd3\xd3\xe3\xbdPS\x8f\x8f\xa7kG\x10G"
        b"\xeb\x87\xfa\x99\x1f\xe0\xb7iG\x9f\x86\xb5\x18\xbb\xb7\x8a]\\|\xfd\xc5\x8e\xe6\xd3\xca}@\\\x0e\x00\x00\x00"
        b"\x000lqt\xf4\x9cC}\x1f}\xf0\x9c#\xcd\x92\xce96\xe8\rm6R\xdf\xb0\x9fs\xb8N\xfbb!\xfc\xbc\x14\xf6Y\x06\xf9:\xa4"
        b"\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
        b"\x00\x00\x00\x00\x1cK>\xde>\xdf\xb9\xf6\xc3\xec\xfc"
        b"}\xf0\xc7\xbe7\xf5m\xe6\xf2\xe7\xa7\xf6^\xee\xc9\xd6\xefG\xb3\x9d\x87\xcf\x9c\xed_\x1f&\xbf\xbe\xd4\xcet\xae"
        b'\xaf\xbf\x99\xc9~\xda\xdd\xaa/\xbf\xb7\xef.\xeen\xbf^2"\xff\x82\xfd\x07\xf7a\x8c\xa2\xd4>\x00\x00'
    )
    import gzip

    fn = get_test_tmp_file(suffix=".hdf")
    with open(fn, "wb") as f:
        f.write(gzip.decompress(raw_gzipped))

    dataset = HDFDataset(files=[fn])
    reader = DatasetTestReader(dataset=dataset)
    reader.read_all()
    print("tags:", reader.seq_tags)
    assert len(seq_lens) == reader.num_seqs
    assert reader.seq_tags == ["seq-0", "seq-1"]
    for i, seq_len in enumerate(seq_lens):
        assert reader.seq_lens[i]["data"] == seq_len
    assert "data" in reader.data_keys  # "classes" might be in there as well, although not really correct/existing
    assert reader.data_sparse["data"] is False
    assert list(reader.data_shape["data"]) == [n_dim]
    assert reader.data_dtype["data"] == "float32"


def test_SimpleHDFWriter_ndim1_var_len():
    # E.g. attention weights, shape (dec-time,enc-time) per seq.
    fn = get_test_tmp_file(suffix=".hdf")
    os.remove(fn)  # SimpleHDFWriter expects that the file does not exist
    writer = SimpleHDFWriter(filename=fn, dim=None, ndim=2, labels=None)
    dec_seq_lens1 = [11, 7, 5]
    enc_seq_lens1 = [13, 6, 8]
    batch1_data = numpy.random.normal(size=(len(dec_seq_lens1), max(dec_seq_lens1), max(enc_seq_lens1))).astype(
        "float32"
    )
    writer.insert_batch(
        inputs=batch1_data,
        seq_len={0: dec_seq_lens1, 1: enc_seq_lens1},
        seq_tag=["seq-%i" % i for i in range(len(dec_seq_lens1))],
    )
    dec_seq_lens2 = [10, 13, 3, 2]
    enc_seq_lens2 = [11, 13, 5, 4]
    batch2_data = numpy.random.normal(size=(len(dec_seq_lens2), max(dec_seq_lens2), max(enc_seq_lens2))).astype(
        "float32"
    )
    writer.insert_batch(
        inputs=batch2_data,
        seq_len={0: dec_seq_lens2, 1: enc_seq_lens2},
        seq_tag=["seq-%i" % (i + len(dec_seq_lens1)) for i in range(len(dec_seq_lens2))],
    )
    writer.close()
    dec_seq_lens = dec_seq_lens1 + dec_seq_lens2
    enc_seq_lens = enc_seq_lens1 + enc_seq_lens2

    dataset = HDFDataset(files=[fn])
    reader = DatasetTestReader(dataset=dataset)
    reader.read_all()
    assert "data" in reader.data_keys  # "classes" might be in there as well, although not really correct/existing
    assert reader.data_sparse["data"] is False
    assert list(reader.data_shape["data"]) == []
    assert reader.data_dtype["data"] == "float32"
    assert "sizes" in reader.data_keys
    # sizes sparsity does not really matter...
    assert reader.data_dtype["sizes"] == "int32"
    assert list(reader.data_shape["sizes"]) == []
    assert len(dec_seq_lens) == len(enc_seq_lens) == reader.num_seqs
    for i, (dec_seq_len, enc_seq_len) in enumerate(zip(dec_seq_lens, enc_seq_lens)):
        assert reader.seq_lens[i]["data"] == dec_seq_len * enc_seq_len
        assert reader.seq_lens[i]["sizes"] == 2
        assert reader.data["sizes"][i].tolist() == [dec_seq_len, enc_seq_len], "got %r" % (reader.data["sizes"][i],)


def test_SimpleHDFWriter_extend_existing_file():
    fn = get_test_tmp_file(suffix=".hdf")
    os.remove(fn)  # SimpleHDFWriter expects that the file does not exist
    n_dim = 3
    writer = SimpleHDFWriter(filename=fn, dim=n_dim, labels=None)
    seq_lens = [2, 3]
    writer.insert_batch(
        inputs=numpy.random.normal(size=(len(seq_lens), max(seq_lens), n_dim)).astype("float32"),
        seq_len=seq_lens,
        seq_tag=["seq-%i" % i for i in range(len(seq_lens))],
    )
    writer.close()
    assert os.path.exists(fn)

    writer = SimpleHDFWriter(filename=fn, dim=n_dim, labels=None, extend_existing_file=True)
    seq_lens2 = [4, 3, 2]
    writer.insert_batch(
        inputs=numpy.random.normal(size=(len(seq_lens2), max(seq_lens2), n_dim)).astype("float32"),
        seq_len=seq_lens2,
        seq_tag=["seq-%i" % (i + len(seq_lens)) for i in range(len(seq_lens2))],
    )
    writer.close()
    seq_lens += seq_lens2

    dataset = HDFDataset(files=[fn])
    reader = DatasetTestReader(dataset=dataset)
    reader.read_all()
    assert "data" in reader.data_keys  # "classes" might be in there as well, although not really correct/existing
    assert reader.data_sparse["data"] is False
    assert list(reader.data_shape["data"]) == [n_dim]
    assert reader.data_dtype["data"] == "float32"
    assert len(seq_lens) == reader.num_seqs
    for i, seq_len in enumerate(seq_lens):
        assert reader.seq_lens[i]["data"] == seq_len


@unittest.skip("unfinished...")
def test_SimpleHDFWriter_swmr():
    fn = get_test_tmp_file(suffix=".hdf")
    os.remove(fn)  # SimpleHDFWriter expects that the file does not exist
    rnd = numpy.random.RandomState(42)
    n_dim = 13
    writer = SimpleHDFWriter(filename=fn, dim=n_dim, labels=None, swmr=True)

    # TODO
    # As we directly want to read it, we must know all the seqs (and seq tags) in advance.
    # Not only that; we also would need all seq lens in advance, to have the offsets in the HDF.
    # (Or do we? We could extend MetaDataset to ignore missing.
    #  But then we would need to enforce a reload at the next epoch. Which we can do.)
    # In any case though, this would not support overwriting existing seqs,
    # but we need that...
    # How to solve this? Different HDF format? But not really possible even in principle in a nice way, or is it?
    # Or write to new HDF file for each epoch?
    # Once finished through whole dataset (all sub epochs), delete old and replace?
    # Then we also don't really need swmr.

    reader = HDFDataset(files=[fn])  # TODO use swmr?
    reader.init_seq_order(epoch=1)

    seq_lens = [11, 7, 5, 10, 13, 3, 2]
    # Construct seq tags in a way that the lengths vary.
    seq_tags = [
        ("seq-%i-" % i) + "".join(["abcdefghijk"[rnd.randint(0, 10)] for _ in range(rnd.randint(1, 50))])
        for i in range(len(seq_lens))
    ]
    seqs = numpy.random.normal(size=(len(seq_lens), max(seq_lens), n_dim)).astype("float32")
    for s, e in [(0, 3), (3, len(seq_lens))]:
        writer.insert_batch(inputs=seqs[s:e, : max(seq_lens[s:e])], seq_len=seq_lens[s:e], seq_tag=seq_tags[s:e])
        writer.flush()  # TODO when do we want it?

        # Now read.
        for seq_idx in range(s, e):
            reader.load_seqs(seq_idx, seq_idx + 1)
            # TODO read and check whether it is correct

    writer.close()  # Should not matter.


def test_SimpleHDFWriter_labels():
    fn = get_test_tmp_file(suffix=".hdf")
    os.remove(fn)  # SimpleHDFWriter expects that the file does not exist
    n_dim = 3
    writer = SimpleHDFWriter(filename=fn, dim=n_dim, labels=[" ", "a", "b"])
    seq_lens = [2, 3]
    writer.insert_batch(
        inputs=numpy.random.normal(size=(len(seq_lens), max(seq_lens), n_dim)).astype("float32"),
        seq_len=seq_lens,
        seq_tag=["seq-%i" % i for i in range(len(seq_lens))],
    )
    writer.close()

    old_dataset = Old2018HDFDataset()
    old_dataset.add_file(fn)
    old_dataset.initialize()
    print("Old dataset:", old_dataset)
    print("Old dataset outputs:", old_dataset.num_outputs)
    print("Old dataset target keys:", old_dataset.target_keys)
    old_dataset.init_seq_order(epoch=1)
    old_dataset.load_seqs(0, 1)

    dataset = HDFDataset(files=[fn])
    print("Dataset:")
    print("  input:", dataset.num_inputs, "x", dataset.window)
    print("  output:", dataset.num_outputs)
    print(" ", dataset.len_info(fast=True) or "no info")
    print("Dataset keys:", dataset.get_data_keys())
    print("Dataset target keys:", dataset.get_target_list())
    print("Dataset labels:", ", ".join(f"{k!r}: {v[:3]}... len {len(v)}" for k, v in dataset.labels.items()) or "None")
    assert dataset.get_data_keys() == ["data"]
    assert dataset.get_target_list() == []

    assert dataset.labels["data"] == [" ", "a", "b"]
    reader = DatasetTestReader(dataset=dataset)
    reader.read_all()
    assert reader.data_keys == ["data"]
    assert reader.data_sparse["data"] is False
    assert list(reader.data_shape["data"]) == [n_dim]
    assert reader.data_dtype["data"] == "float32"
    assert len(seq_lens) == reader.num_seqs
    for i, seq_len in enumerate(seq_lens):
        assert reader.seq_lens[i]["data"] == seq_len


class Old2018HDFDataset(CachedDataset):
    """
    Copied and adapted from an early RETURNN version:
    2018-03-09: https://github.com/rwth-i6/returnn/blob/c2d8fed877022d1ac1bf68b801604733db51223e/HDFDataset.py

    (Not really fully functional here though... we just use it to test whether loading works.)

    Some version history:
    2015-01-05: https://github.com/rwth-i6/returnn/blob/3be0fe0906212d1ffdd93807c0a3854a38842eb8/Dataset.py
    2015-08-04: https://github.com/rwth-i6/returnn/blob/995e87184abc6e07256417cb533163ac0a7d7dd8/HDFDataset.py
    2018-03-09: https://github.com/rwth-i6/returnn/blob/c2d8fed877022d1ac1bf68b801604733db51223e/HDFDataset.py
    """

    def __init__(self, **kwargs):
        super(Old2018HDFDataset, self).__init__(cache_byte_size=100, **kwargs)
        self.files: List[str] = []
        self.file_start = [0]
        self.file_seq_start: List[List[int]] = []
        self.file_index: List[int] = []
        self.data_dtype: Dict[str, str] = {}
        self.data_sparse: Dict[str, bool] = {}

        # Copied from old base CachedDataset
        self._seq_lengths: List[Tuple[int, int]] = []  # uses real seq idx
        self.tags: List[str] = []  # uses real seq idx
        self.tag_idx: Dict[str, int] = {}  # map of tag -> real-seq-idx
        self.targets = {}
        self.target_keys = []

        # Copied from old base Dataset
        self._num_codesteps: Optional[int] = None  # Num output frames, could be different from input, seq2seq, ctc.

    def add_file(self, filename):
        """
        Setups data:
          self.seq_lengths
          self.file_index
          self.file_start
          self.file_seq_start
        Use load_seqs() to load the actual data.
        :type filename: str
        """
        fin = h5py.File(filename, "r")
        decode = lambda s: s if isinstance(s, str) else s.decode("utf-8")
        if "targets" in fin:
            self.labels = {
                k: [decode(item).split("\0")[0] for item in fin["targets/labels"][k][...].tolist()]
                for k in fin["targets/labels"]
            }
        if not self.labels:
            labels = [decode(item).split("\0")[0] for item in fin["labels"][...].tolist()]
            """:type: list[str]"""
            self.labels = {"classes": labels}
            assert len(self.labels["classes"]) == len(labels), (
                "expected " + str(len(self.labels["classes"])) + " got " + str(len(labels))
            )
        tags = [decode(item).split("\0")[0] for item in fin["seqTags"][...].tolist()]
        """ :type: list[str] """
        self.files.append(filename)
        if "times" in fin:
            if self.timestamps is None:
                self.timestamps = fin[attr_times][...]
            else:
                self.timestamps = numpy.concatenate(
                    [self.timestamps, fin[attr_times][...]], axis=0
                )  # .extend(fin[attr_times][...].tolist())
        seq_lengths = fin[attr_seqLengths][...]
        if "targets" in fin:
            self.target_keys = sorted(fin["targets/labels"].keys())
        else:
            self.target_keys = ["classes"]

        if len(seq_lengths.shape) == 1:
            seq_lengths = numpy.array(zip(*[seq_lengths.tolist() for i in range(len(self.target_keys) + 1)]))

        seq_start = [numpy.zeros((seq_lengths.shape[1],), "int64")]
        if not self._seq_start:
            self._seq_start = [numpy.zeros((seq_lengths.shape[1],), "int64")]
        for l in seq_lengths:
            self._seq_lengths.append(numpy.array(l))
            seq_start.append(seq_start[-1] + l)
        self.tags += tags
        self.file_seq_start.append(seq_start)
        nseqs = len(seq_start) - 1
        for i in range(nseqs):
            self.tag_idx[tags[i]] = i + self._num_seqs
        self._num_seqs += nseqs
        self.file_index.extend([len(self.files) - 1] * nseqs)
        self.file_start.append(self.file_start[-1] + nseqs)
        self._num_timesteps += sum([s[0] for s in seq_lengths])
        if self._num_codesteps is None:
            self._num_codesteps = [0 for i in range(1, len(seq_lengths[0]))]
        for i in range(1, len(seq_lengths[0])):
            self._num_codesteps[i - 1] += sum([s[i] for s in seq_lengths])
        if len(fin["inputs"].shape) == 1:  # sparse
            num_inputs = [fin.attrs[attr_inputPattSize], 1]
        else:
            num_inputs = [fin["inputs"].shape[1], len(fin["inputs"].shape)]  # fin.attrs[attr_inputPattSize]
        if self.num_inputs == 0:
            self.num_inputs = num_inputs[0]
        assert self.num_inputs == num_inputs[0], "wrong input dimension in file %s (expected %s got %s)" % (
            filename,
            self.num_inputs,
            num_inputs[0],
        )
        if "targets/size" in fin:
            num_outputs = {
                k: [fin["targets/size"].attrs[k], len(fin["targets/data"][k].shape)] for k in fin["targets/size"].attrs
            }
        else:
            num_outputs = {"classes": fin.attrs["numLabels"]}
        num_outputs["data"] = num_inputs
        if not self.num_outputs:
            self.num_outputs = num_outputs
        assert self.num_outputs == num_outputs, "wrong dimensions in file %s (expected %s got %s)" % (
            filename,
            self.num_outputs,
            num_outputs,
        )
        if "targets" in fin:
            for name in fin["targets/data"]:
                tdim = 1 if len(fin["targets/data"][name].shape) == 1 else fin["targets/data"][name].shape[1]
                self.data_dtype[name] = str(fin["targets/data"][name].dtype) if tdim > 1 else "int32"
                self.targets[name] = None
        else:
            self.targets = {"classes": numpy.zeros((self._num_timesteps,))}
            self.data_dtype["classes"] = "int32"
        self.data_dtype["data"] = fin["inputs"].dtype
        assert len(self.target_keys) == len(self._seq_lengths[0]) - 1
        fin.close()

    def _load_seqs(self, start, end):
        """
        Load data sequences.
        As a side effect, will modify / fill-up:
          self.alloc_intervals
          self.targets
          self.chars

        :param int start: start sorted seq idx
        :param int end: end sorted seq idx
        """
        assert start < self.num_seqs
        assert end <= self.num_seqs
        selection = self.insert_alloc_interval(start, end)
        assert len(selection) <= end - start, (
            "DEBUG: more sequences requested (" + str(len(selection)) + ") as required (" + str(end - start) + ")"
        )
        file_info: List[List[int]] = [[] for l in range(len(self.files))]
        # file_info[i] is (sorted seq idx from selection, real seq idx)
        for idc in selection:
            ids = self._seq_index[idc]
            file_info[self.file_index[ids]].append((idc, ids))
            self.preload_set.add(ids)
        for i in range(len(self.files)):
            if len(file_info[i]) == 0:
                continue
            print("loading file %d/%d" % (i + 1, len(self.files)), self.files[i], file=log.v4)
            fin = h5py.File(self.files[i], "r")
            for idc, ids in file_info[i]:
                s = ids - self.file_start[i]
                p = self.file_seq_start[i][s]
                l = self._seq_lengths[ids]
                if "targets" in fin:
                    for k in fin["targets/data"]:
                        if self.targets[k] is None:
                            self.targets[k] = numpy.zeros((self._num_codesteps[self.target_keys.index(k)],)) - 1
                        ldx = self.target_keys.index(k) + 1
                        self.targets[k][self.get_seq_start(idc)[ldx] : self.get_seq_start(idc)[ldx] + l[ldx]] = fin[
                            "targets/data/" + k
                        ][p[ldx] : p[ldx] + l[ldx]]
                self._set_alloc_intervals_data(idc, data=fin["inputs"][p[0] : p[0] + l[0]][...])
            fin.close()
        gc.collect()
        assert self.is_cached(start, end)

    def _get_seq_length_by_real_idx(self, real_seq_idx):
        return self._seq_lengths[real_seq_idx]

    def get_tag(self, sorted_seq_idx):
        ids = self._seq_index[self._index_map[sorted_seq_idx]]
        return self.tags[ids]

    def is_data_sparse(self, key):
        if key in self.num_outputs:
            return self.num_outputs[key][1] == 1
        if self.get_data_dtype(key).startswith("int"):
            return True
        return False

    def get_data_dtype(self, key):
        return self.data_dtype[key]

    def len_info(self):
        return ", ".join(["HDF dataset", "sequences: %i" % self.num_seqs, "frames: %i" % self.get_num_timesteps()])


def dummy_iter_dataset(dataset: Dataset) -> int:
    """
    :param Dataset dataset:
    :return: num seqs
    """
    dataset.init_seq_order(epoch=1)
    data_keys = dataset.get_data_keys()
    seq_idx = 0
    while dataset.is_less_than_num_seqs(seq_idx):
        dataset.load_seqs(seq_idx, seq_idx + 1)
        for key in data_keys:
            dataset.get_data(seq_idx=seq_idx, key=key)
            dataset.get_tag(seq_idx)
        seq_idx += 1
    print("Iterated through %r, num seqs %i" % (dataset, seq_idx))
    return seq_idx


def test_hdf_simple_iter():
    hdf_fn = generate_hdf_from_dummy()
    dataset = HDFDataset(files=[hdf_fn])
    dataset.initialize()
    dummy_iter_dataset(dataset)


def test_hdf_simple_iter_cached():
    hdf_fn = generate_hdf_from_dummy()
    dataset = HDFDataset(files=[hdf_fn], cache_byte_size=100)
    dataset.initialize()
    dummy_iter_dataset(dataset)


def test_rnn_getCacheByteSizes_zero():
    from returnn.config import Config

    config = Config({"cache_size": "0"})
    import returnn.__main__ as rnn

    rnn.config = config
    sizes = rnn.get_cache_byte_sizes()
    assert len(sizes) == 3
    assert all([s == 0 for s in sizes])


def test_rnn_initData():
    hdf_fn = generate_hdf_from_dummy()
    from returnn.config import Config

    config = Config({"cache_size": "0", "train": hdf_fn, "dev": hdf_fn})
    import returnn.__main__ as rnn

    rnn.config = config
    rnn.init_data()
    train, dev = rnn.train_data, rnn.dev_data
    assert train and dev
    assert isinstance(train, HDFDataset)
    assert isinstance(dev, HDFDataset)
    assert train.cache_byte_size_total_limit == dev.cache_byte_size_total_limit == 0
    assert train.cache_byte_size_limit_at_start == dev.cache_byte_size_limit_at_start == 0


def test_hdf_no_cache_iter():
    hdf_fn = generate_hdf_from_dummy()
    dataset = HDFDataset(files=[hdf_fn])
    dataset.initialize()
    assert dataset.cache_byte_size_limit_at_start == 0
    assert dataset.cache_byte_size_total_limit == 0

    class DummyCallback:
        def __init__(self):
            self.was_called = False

        def __call__(self, *args, **kwargs):
            print("DummyCallback called!")
            self.was_called = True

    assert hasattr(dataset, "_preload_seqs")
    dataset._preload_seqs = DummyCallback()

    dummy_iter_dataset(dataset)
    import time

    time.sleep(1)  # maybe threads are running in background...
    assert not dataset._preload_seqs.was_called


def test_hdf_data_short_int_dtype():
    from returnn.datasets.generating import StaticDataset

    dataset = StaticDataset(
        [{"data": numpy.array([1, 2, 3], dtype="uint8"), "classes": numpy.array([-1, 5], dtype="int16")}],
        output_dim={"data": (255, 1), "classes": (10, 1)},
    )
    orig_data_dtype = dataset.get_data_dtype("data")
    orig_classes_dtype = dataset.get_data_dtype("classes")
    assert orig_data_dtype == "uint8" and orig_classes_dtype == "int16"

    hdf_fn = get_test_tmp_file(suffix=".hdf")
    hdf_writer = HDFDatasetWriter(filename=hdf_fn)
    hdf_writer.dump_from_dataset(dataset, use_progress_bar=False)
    hdf_writer.close()

    hdf_dataset = HDFDataset(files=[hdf_fn])
    hdf_dataset.initialize()
    hdf_dataset.init_seq_order(epoch=1)
    hdf_data_dtype = hdf_dataset.get_data_dtype("data")
    hdf_classes_dtype = hdf_dataset.get_data_dtype("classes")
    assert hdf_data_dtype == orig_data_dtype and hdf_classes_dtype == orig_classes_dtype
    hdf_data_dim = hdf_dataset.get_data_dim("data")
    hdf_classes_dim = hdf_dataset.get_data_dim("classes")
    assert hdf_data_dim == 255 and hdf_classes_dim == 10
    hdf_data_shape = hdf_dataset.get_data_shape("data")
    hdf_classes_shape = hdf_dataset.get_data_shape("classes")
    assert hdf_data_shape == [] and hdf_classes_shape == []
    hdf_dataset.load_seqs(0, 1)
    hdf_data_data = hdf_dataset.get_data(0, "data")
    hdf_data_classes = hdf_dataset.get_data(0, "classes")
    assert hdf_data_data.dtype == orig_data_dtype and hdf_data_classes.dtype == orig_classes_dtype


def test_hdf_data_target_int32():
    from returnn.datasets.generating import StaticDataset

    dataset = StaticDataset(
        [
            {
                "data": numpy.array([1, 2, 3], dtype="uint8"),
                "classes": numpy.array([2147483647, 2147483646, 2147483645], dtype="int32"),
            }
        ],
        output_dim={"data": (255, 1), "classes": (10, 1)},
    )
    dataset.initialize()
    dataset.init_seq_order(epoch=0)
    dataset.load_seqs(0, 1)
    orig_classes_dtype = dataset.get_data_dtype("classes")
    orig_classes_seq = dataset.get_data(0, "classes")
    assert orig_classes_seq.shape == (3,) and orig_classes_seq[0] == 2147483647
    assert orig_classes_seq.dtype == orig_classes_dtype == "int32"

    hdf_fn = get_test_tmp_file(suffix=".hdf")
    hdf_writer = HDFDatasetWriter(filename=hdf_fn)
    hdf_writer.dump_from_dataset(dataset, use_progress_bar=False)
    hdf_writer.close()

    hdf_dataset = HDFDataset(files=[hdf_fn])
    hdf_dataset.initialize()
    hdf_dataset.init_seq_order(epoch=1)
    hdf_classes_dtype = hdf_dataset.get_data_dtype("classes")
    assert hdf_classes_dtype == orig_classes_dtype
    hdf_classes_shape = hdf_dataset.get_data_shape("classes")
    assert hdf_classes_shape == []
    hdf_dataset.load_seqs(0, 1)
    hdf_data_classes = hdf_dataset.get_data(0, "classes")
    assert hdf_data_classes.dtype == orig_classes_dtype
    assert all(hdf_data_classes == orig_classes_seq)


def test_hdf_target_float_dtype():
    from returnn.datasets.generating import StaticDataset

    dataset = StaticDataset(
        [{"data": numpy.array([1, 2, 3], dtype="float32"), "classes": numpy.array([-1, 5], dtype="float32")}],
        output_dim={"data": (1, 1), "classes": (1, 1)},
    )
    orig_data_dtype = dataset.get_data_dtype("data")
    orig_classes_dtype = dataset.get_data_dtype("classes")
    assert orig_data_dtype == "float32" and orig_classes_dtype == "float32"

    hdf_fn = get_test_tmp_file(suffix=".hdf")
    hdf_writer = HDFDatasetWriter(filename=hdf_fn)
    hdf_writer.dump_from_dataset(dataset, use_progress_bar=False)
    hdf_writer.close()

    hdf_dataset = HDFDataset(files=[hdf_fn])
    hdf_dataset.initialize()
    hdf_dataset.init_seq_order(epoch=1)
    hdf_data_dtype = hdf_dataset.get_data_dtype("data")
    hdf_classes_dtype = hdf_dataset.get_data_dtype("classes")
    assert hdf_data_dtype == orig_data_dtype and hdf_classes_dtype == orig_classes_dtype
    hdf_data_dim = hdf_dataset.get_data_dim("data")
    hdf_classes_dim = hdf_dataset.get_data_dim("classes")
    assert hdf_data_dim == 1 and hdf_classes_dim == 1
    hdf_data_shape = hdf_dataset.get_data_shape("data")
    hdf_classes_shape = hdf_dataset.get_data_shape("classes")
    assert hdf_data_shape == [] and hdf_classes_shape == []
    hdf_dataset.load_seqs(0, 1)
    hdf_data_data = hdf_dataset.get_data(0, "data")
    hdf_data_classes = hdf_dataset.get_data(0, "classes")
    assert hdf_data_data.dtype == orig_data_dtype and hdf_data_classes.dtype == orig_classes_dtype


def test_hdf_target_float_dense():
    from returnn.datasets.generating import StaticDataset

    dataset = StaticDataset(
        [
            {
                "data": numpy.array([[1, 2, 3], [2, 3, 4]], dtype="float32"),
                "classes": numpy.array([[-1, 5], [-2, 4], [-3, 2]], dtype="float32"),
            }
        ]
    )
    orig_data_dtype = dataset.get_data_dtype("data")
    orig_classes_dtype = dataset.get_data_dtype("classes")
    assert orig_data_dtype == "float32" and orig_classes_dtype == "float32"
    orig_data_shape = dataset.get_data_shape("data")
    orig_classes_shape = dataset.get_data_shape("classes")
    assert orig_data_shape == [3] and orig_classes_shape == [2]

    hdf_fn = get_test_tmp_file(suffix=".hdf")
    hdf_writer = HDFDatasetWriter(filename=hdf_fn)
    hdf_writer.dump_from_dataset(dataset, use_progress_bar=False)
    hdf_writer.close()

    hdf_dataset = HDFDataset(files=[hdf_fn])
    hdf_dataset.initialize()
    hdf_dataset.init_seq_order(epoch=1)
    hdf_data_dtype = hdf_dataset.get_data_dtype("data")
    hdf_classes_dtype = hdf_dataset.get_data_dtype("classes")
    assert hdf_data_dtype == orig_data_dtype and hdf_classes_dtype == orig_classes_dtype
    hdf_data_dim = hdf_dataset.get_data_dim("data")
    hdf_classes_dim = hdf_dataset.get_data_dim("classes")
    assert hdf_data_dim == orig_data_shape[-1] and hdf_classes_dim == orig_classes_shape[-1]
    hdf_data_shape = hdf_dataset.get_data_shape("data")
    hdf_classes_shape = hdf_dataset.get_data_shape("classes")
    assert hdf_data_shape == orig_data_shape and hdf_classes_shape == orig_classes_shape
    hdf_dataset.load_seqs(0, 1)
    hdf_data_data = hdf_dataset.get_data(0, "data")
    hdf_data_classes = hdf_dataset.get_data(0, "classes")
    assert hdf_data_data.dtype == orig_data_dtype and hdf_data_classes.dtype == orig_classes_dtype


def test_HDFDataset_no_cache_efficiency():
    hdf_fn = generate_hdf_from_other({"class": "Task12AXDataset", "num_seqs": 23})
    hdf_dataset = HDFDataset(files=[hdf_fn], cache_byte_size=0)
    hdf_dataset.initialize()
    hdf_dataset.init_seq_order(epoch=1)
    hdf_dataset.load_seqs(0, 1)
    hdf_dataset.load_seqs(0, 2)
    hdf_dataset.load_seqs(0, 5)
    hdf_dataset.load_seqs(1, 2)
    hdf_dataset.load_seqs(1, 3)
    hdf_dataset.load_seqs(1, 7)
    hdf_dataset.load_seqs(3, 7)
    hdf_dataset.load_seqs(4, 10)
    # TODO... check alloc intervals etc


def test_HDFDataset_pickle():
    hdf_fn = generate_hdf_from_other({"class": "Task12AXDataset", "num_seqs": 23})
    ds = HDFDataset(files=[hdf_fn], cache_byte_size=0)
    ds.initialize()
    ds.init_seq_order(epoch=1)

    import pickle

    ds_copy = pickle.loads(pickle.dumps(ds))
    assert ds_copy.epoch == 1
    num_seqs = dummy_iter_dataset(ds)
    num_seqs_copy = dummy_iter_dataset(ds_copy)
    assert num_seqs == num_seqs_copy


def test_HDFDataset_deepcopy():
    hdf_fn = generate_hdf_from_other({"class": "Task12AXDataset", "num_seqs": 23})
    ds = HDFDataset(files=[hdf_fn], cache_byte_size=0)
    ds.initialize()
    ds.init_seq_order(epoch=1)

    import copy

    ds_copy = copy.deepcopy(ds)
    assert ds_copy.epoch == 1
    num_seqs = dummy_iter_dataset(ds)
    num_seqs_copy = dummy_iter_dataset(ds_copy)
    assert num_seqs == num_seqs_copy


def test_siamese_triplet_sampling():
    datasets_path = generate_dummy_hdf(3)
    dataset = SiameseHDFDataset(input_stream_name="features", seq_label_stream="classes", files=datasets_path)

    dataset.initialize()
    for iter in range(1, 31):
        print("Initializing triplets... iteration %d" % iter)
        dataset.init_seq_order(epoch=iter)

        triplets = dataset.curr_epoch_triplets
        anchor_seq_names = [dataset.all_seq_names[id[0]] for id in triplets]
        same_class_seq_names = [dataset.all_seq_names[id[1]] for id in triplets]
        diff_class_seq_names = [dataset.all_seq_names[id[2]] for id in triplets]

        anchor_class = [dataset.seq_to_target[seq_id] for seq_id in anchor_seq_names]
        same_class = [dataset.seq_to_target[seq_id] for seq_id in same_class_seq_names]
        diff_class = [dataset.seq_to_target[seq_id] for seq_id in diff_class_seq_names]

        print("Testing pair sequences to belong to the same class...")
        assert all(ac == same_class[id] for id, ac in enumerate(anchor_class))
        print("Testing third element in a triplet to belong to a different class...")
        assert all(ac != diff_class[id] for id, ac in enumerate(anchor_class))
        print("------------------------------------------------------")

    print("Deleting temporary files...")
    for path in datasets_path:
        os.remove(path)
    print("Done.")


def test_siamese_collect_single_seq():
    datasets_path = generate_dummy_hdf(3)
    dataset = SiameseHDFDataset(input_stream_name="features", seq_label_stream="classes", files=datasets_path)

    dataset.initialize()
    dataset.init_seq_order(epoch=1)

    random = np.random.RandomState()
    seq_idx = random.randint(low=0, high=len(dataset.seq_name_to_idx))
    dataset_seq = dataset._collect_single_seq(seq_idx)
    print("Verify that single sequence consists of a triplet...")
    print(dataset_seq.features.keys())

    print("Deleting temporary files...")
    for path in datasets_path:
        os.remove(path)
    print("Done.")


if __name__ == "__main__":
    better_exchook.install()
    if len(sys.argv) <= 1:
        for k, v in sorted(globals().items()):
            if k.startswith("test_"):
                print("-" * 40)
                print("Executing: %s" % k)
                try:
                    v()
                except unittest.SkipTest as exc:
                    print("SkipTest:", exc)
                print("-" * 40)
        print("Finished all tests.")
    else:
        assert len(sys.argv) >= 2
        for arg in sys.argv[1:]:
            print("Executing: %s" % arg)
            if arg in globals():
                globals()[arg]()  # assume function and execute
            else:
                eval(arg)  # assume Python code and execute



"""
Normalization functions such as batch norm
"""

from __future__ import annotations
from typing import Optional, Sequence, Union, Tuple
from returnn.tensor import Tensor, Dim
import returnn.frontend as rf


__all__ = ["moments", "LayerNorm", "RMSNorm", "GroupNorm", "BatchNorm", "normalize", "Normalize"]


def moments(
    x: Tensor, axis: Union[Dim, Sequence[Dim]], *, use_mask: bool = True, correction: Union[int, float, Tensor] = 0
) -> Tuple[Tensor, Tensor]:
    """
    :param x: input
    :param axis: the axis (or axes) to be reduced, to calculate statistics over
    :param use_mask: whether to use a mask for dynamic spatial dims in the reduction
    :param correction:
        The variance will be estimated by ``sum((x - mean)**2) / (n-correction)``
        where ``n`` is the number of elements in the axis (or the axes)
        (with ``use_mask=True``, taking masking into account, using :func:`num_elements_of_shape`).
        The default ``correction=0`` will return the biased variance estimation.
        ``correction=1`` is the `Bessel correction <https://en.wikipedia.org/wiki/Bessel%27s_correction>`__
        and will return the unbiased variance estimation.
        In PyTorch, there was an argument ``unbiased`` for this, but this changed recently to ``correction``
        (`PyTorch issue #61492 <https://github.com/pytorch/pytorch/issues/61492>`__,
         `Python Array API Standard
          <https://data-apis.org/array-api/latest/API_specification/generated/array_api.var.html>`__).
        In PyTorch, the default is ``correction=1``, which is the unbiased variance estimation,
        while in most other frameworks, the default is ``correction=0``, which is the biased variance estimation.
    :return: tuple (mean, variance). it has the same shape as the input with the axis removed
    """
    mean = rf.reduce_mean(x, axis=axis)
    # stop_gradient does not change the gradient here
    variance = rf.reduce_mean(rf.squared_difference(x, rf.stop_gradient(mean)), axis=axis, use_mask=use_mask)
    if isinstance(correction, Tensor) or correction != 0:
        n = rf.num_elements_of_shape(axis, use_mask=use_mask)
        variance *= n / (n - correction)
    return mean, variance


class LayerNorm(rf.Module):
    """
    `Layer normalization <https://arxiv.org/abs/1607.06450>`__.

    Note that we *just* normalize over the feature-dim axis here.
    This is consistent to the default behavior of :class:`tf.keras.layers.LayerNormalization`
    and also how it is commonly used in many models, including Transformer.

    However, there are cases where it would be common to normalize over all axes except batch-dim,
    or all axes except batch and time.
    For a more generic variant, see :func:`norm`.
    """

    def __init__(self, in_dim: Union[rf.Dim, Sequence[rf.Dim]], *, eps: float = 1e-6, with_bias: bool = True):
        super().__init__()
        self.in_dim = in_dim
        self.eps = eps
        self.scale = rf.Parameter([self.in_dim] if isinstance(self.in_dim, rf.Dim) else self.in_dim)
        self.scale.initial = 1.0
        self.bias = None
        if with_bias:
            self.bias = rf.Parameter(self.scale.dims)
            self.bias.initial = 0.0

    def __call__(self, x: Tensor) -> Tensor:
        mean, variance = rf.moments(x, axis=self.in_dim)
        norm_x = (x - mean) * rf.rsqrt(variance + self.eps)
        out = norm_x * self.scale
        if self.bias is not None:
            out += self.bias
        return out


class RMSNorm(rf.Module):
    """
    `Root Mean Square Layer Normalization (RMSNorm) <https://arxiv.org/abs/1910.07467>`__.

    Alternative to :class:`LayerNorm` that uses the root-mean-square of the input as the normalization factor.
    I.e. the main difference to layer norm is: *No subtraction of mean.*

    Note, the bias here is optional, *and disabled by default* (in line with most implementations of RMSNorm),
    unlike our :class:`LayerNorm`, where the bias is enabled by default.
    """

    def __init__(self, in_dim: Union[rf.Dim, Sequence[rf.Dim]], *, eps: float = 1e-6, with_bias: bool = False):
        super().__init__()
        self.in_dim = in_dim
        self.eps = eps
        self.scale = rf.Parameter([self.in_dim] if isinstance(self.in_dim, rf.Dim) else self.in_dim)
        self.scale.initial = 1.0
        self.bias = None
        if with_bias:
            self.bias = rf.Parameter(self.scale.dims)
            self.bias.initial = 0.0

    def __call__(self, x: Tensor) -> Tensor:
        variance = rf.reduce_mean(rf.square(x), axis=self.in_dim)
        norm_x = x * rf.rsqrt(variance + self.eps)
        out = norm_x * self.scale
        if self.bias is not None:
            out += self.bias
        return out


class GroupNorm(rf.Module):
    """
    `Group normalization <https://arxiv.org/abs/1803.08494>`__.
    """

    def __init__(self, in_dim: Union[rf.Dim, Sequence[rf.Dim]], *, num_groups: Union[int, Dim], eps: float = 1e-6):
        super().__init__()
        self.in_dim = in_dim
        self.num_groups = num_groups if isinstance(num_groups, Dim) else Dim(num_groups, name="groups")
        self.in_group_dim = in_dim.ceildiv_left(num_groups)
        self.eps = eps
        self.scale = rf.Parameter([self.in_dim] if isinstance(self.in_dim, rf.Dim) else self.in_dim)
        self.scale.initial = 1.0
        self.bias = rf.Parameter(self.scale.dims)
        self.bias.initial = 0.0

    def __call__(self, x: Tensor) -> Tensor:
        x = rf.split_dims(x, axis=self.in_dim, dims=[self.num_groups, self.in_group_dim])
        mean, variance = rf.moments(x, axis=self.in_group_dim)
        norm_x = (x - mean) * rf.rsqrt(variance + self.eps)
        norm_x, _ = rf.merge_dims(norm_x, dims=[self.num_groups, self.in_group_dim], out_dim=self.in_dim)
        return norm_x * self.scale + self.bias


class BatchNorm(rf.Module):
    """
    Batch normalization. https://arxiv.org/abs/1502.03167

    Note that the default arguments differ from corresponding batch norm in RETURNN.
    See here for discussion on defaults: https://github.com/rwth-i6/returnn/issues/522

    We calculate statistics over all axes except the given in_dim.
    I.e. all other axes are reduced for the statistics.

    To compensate the normalization, there are learnable parameters gamma and beta
    (optional, used when option `affine` is True).

    The usual behavior depends on whether this is used in training or evaluation,
    although this often configurable in other frameworks.
    The usual behavior, in training::

        # Using statistics from current batch.
        mean_cur_batch, variance_cur_batch = moments(source, reduce_dims)
        y = (x - mean_cur_batch) / sqrt(variance_cur_batch + epsilon)
        y = gamma * y + beta

        # Updating running statistics for later use.
        mean = (1 - momentum) * mean + momentum * mean_cur_batch
        variance = (1 - momentum) * variance + momentum * variance_cur_batch

    The usual behavior, not in training (i.e. in evaluation)::

        # Using collected statistics. Not using statistics from current batch.
        y = (x - mean) / sqrt(variance + epsilon)
        y = gamma * y + beta

    """

    def __init__(
        self,
        in_dim: Dim,
        *,
        affine: bool = True,
        momentum: float = 0.1,
        eps: float = 1e-3,
        track_running_stats: bool = True,
        use_mask: Optional[bool] = None,
    ):
        """
        :param in_dim: the feature dimension of the input
        :param affine: whether to use learnable parameters gamma and beta
        :param momentum: momentum for the running mean and variance
        :param eps: epsilon for the variance
        :param track_running_stats:
            If True, uses statistics of the current batch for normalization during training,
            and the tracked statistics (running mean and variance) during evaluation.
            If False, uses statistics of the current batch for normalization during both training and evaluation.
        :param use_mask: whether to use a mask for dynamic spatial dims.
          This must be specified if the input has dynamic spatial dims.
          True would use the correct masking then. However, that is inconsistent to all other frameworks
            which ignore the masking, and also slower, and the fused op would not be used.
          False would be consistent to all other frameworks,
            and potentially allows for the use of an efficient fused op internally.
        """
        super().__init__()
        assert isinstance(in_dim, Dim)
        self.in_dim = in_dim
        self.use_mask = use_mask
        self.momentum = momentum
        self.eps = eps
        if track_running_stats:
            self.running_mean = rf.Parameter([in_dim], auxiliary=True)
            self.running_mean.initial = 0.0
            self.running_variance = rf.Parameter([in_dim], auxiliary=True)
            self.running_variance.initial = 1.0
        else:
            self.running_mean = None
            self.running_variance = None
        self.affine = affine
        self.gamma = None  # type: Optional[rf.Parameter]
        self.beta = None  # type: Optional[rf.Parameter]
        if self.affine:
            self.gamma = rf.Parameter([in_dim])
            self.gamma.initial = 1.0
            self.beta = rf.Parameter([in_dim])
            self.beta.initial = 0.0

    def __call__(self, source: Tensor) -> Tensor:
        assert self.in_dim in source.dims

        if any(d.need_masking() for d in source.dims if d != self.in_dim):
            if self.use_mask is None:
                raise ValueError(
                    f"{self}: use_mask must be specified if the input {source} has any dynamic spatial dims"
                )
            use_mask = self.use_mask
        else:
            use_mask = False  # not needed. False because this potentially enables an efficient fused op.

        if use_mask:
            # Generic implementation which supports masking.
            use_current_batch_stats = self.running_mean is None or rf.get_run_ctx().train_flag
            update_running_stats = self.running_mean is not None and rf.get_run_ctx().train_flag
            need_current_batch_stats = rf.opt_logical_or(use_current_batch_stats, update_running_stats)

            mean_cur_batch, variance_cur_batch = rf.cond(
                need_current_batch_stats,
                lambda: rf.moments(source, axis=[d for d in source.dims if d != self.in_dim]),
                lambda: (self.running_mean, self.running_variance),
            )

            def _update_running_stats():
                self.running_mean.assign_add((mean_cur_batch - self.running_mean) * self.momentum)
                self.running_variance.assign_add((variance_cur_batch - self.running_variance) * self.momentum)

            rf.cond(update_running_stats, _update_running_stats, lambda: None)

            mean, variance = rf.cond(
                use_current_batch_stats,
                lambda: (mean_cur_batch, variance_cur_batch),
                lambda: (self.running_mean, self.running_variance),
            )

            bn = (source - mean) * rf.rsqrt(variance + self.eps)
            if self.gamma is not None:
                bn *= self.gamma
            if self.beta is not None:
                bn += self.beta
            return bn

        # Fallback to specific backend implementation for the standard case without masking.
        # This fallback probably can internally use a more efficient implementation like from CuDNN.
        # In case of TF-net-dict backend, we wrap the RETURNN layer because we want efficient handling if possible,
        # which is potentially the use of a fused op,
        # and maybe reordering of dims.
        # https://github.com/rwth-i6/returnn_common/issues/89
        # noinspection PyProtectedMember
        return source._raw_backend.batch_norm(
            source=source,
            in_dim=self.in_dim,
            use_mask=use_mask,
            affine=self.affine,
            momentum=self.momentum,
            epsilon=self.eps,
            running_mean=self.running_mean,
            running_variance=self.running_variance,
            gamma=self.gamma,
            beta=self.beta,
        )


def normalize(a: Tensor, *, axis: Union[Dim, Sequence[Dim]], epsilon: float = 1e-6) -> Tensor:
    """
    Mean- and variance-normalize some input in the given input dimension(s),
    such that the resulting tensor has mean 0 and variance 1.

    If you want that this can be shifted and scaled again,
    you need additional parameters, cf. :class:`Normalize`.

    :param a: input
    :param axis: axis over which the mean and variance are computed
    :param epsilon: epsilon for numerical stability
    :return: (a - mean) / sqrt(variance + epsilon)
    """
    mean, variance = rf.moments(a, axis=axis)
    return (a - mean) * rf.rsqrt(variance + epsilon)


class Normalize(rf.Module):
    """
    :func:`normalize` with additional scale and bias
    """

    def __init__(
        self,
        *,
        param_dims: Union[Dim, Sequence[Dim]],
        epsilon: float = 1e-6,
        scale: bool = True,
        bias: bool = True,
    ):
        """
        :param param_dims: shape of the scale and bias parameters
        :param epsilon: epsilon for numerical stability
        :param scale: whether to include a trainable scale
        :param bias: whether to include a trainable bias
        """
        super(Normalize, self).__init__()
        self.epsilon = epsilon
        if isinstance(param_dims, Dim):
            param_dims = [param_dims]
        self.scale = None
        if scale:
            self.scale = rf.Parameter(dims=param_dims)
            self.scale.initial = 1.0
        self.bias = rf.Parameter(dims=param_dims) if bias else None

    def __call__(self, a: Tensor, *, axis: Union[Dim, Sequence[Dim]]):
        norm = normalize(a, axis=axis, epsilon=self.epsilon)
        if self.scale is not None:
            norm = self.scale * norm
        if self.bias is not None:
            norm = norm + self.bias
        return norm



import _setup_test_env  # noqa
import tensorflow as tf
import sys
import os
import contextlib
import unittest
import numpy.testing
from pprint import pprint
from returnn.util import better_exchook
from returnn.config import Config
from returnn.tf.network import *
from returnn.tf.layers.basic import *
from returnn.tf.engine import *
from returnn.log import log
import returnn.tf.compat as tf_compat
import returnn.tf.util.basic as tf_util


@contextlib.contextmanager
def make_scope():
    with tf.Graph().as_default() as graph:
        with tf_compat.v1.Session(graph=graph) as session:
            yield session


network = {}
_last = "data"


def build_resnet(conv_time_dim):
    # network
    # (also defined by num_inputs & num_outputs)
    dropout = 0
    L2 = 0.1
    filter_size = (3, 3)  # for 2D conv on (window, feature) axes

    # data
    context_window = 1

    window = 1
    feature_dim = 64  # Gammatone 40-dim
    channel_num = 3
    num_inputs = feature_dim * channel_num * window
    num_outputs = 9001  # CART labels
    EpochSplit = 6

    cur_feat_dim = feature_dim

    global _last, network
    network = {}
    _last = "data"

    def add_sequential_layer(name, d, from_=None):
        global _last, network
        assert "from" not in d
        if from_ is not None:
            d["from"] = from_
        else:
            d["from"] = [_last]
        assert name not in network
        network[name] = d
        _last = name

    def fixed_padding(prefix, kernel_size, data_format, conv_time_dim):
        """Pads the input along the spatial dimensions independently of input size."""
        pad_total = kernel_size - 1
        feature_pad_beg = pad_total // 2
        feature_pad_end = pad_total - feature_pad_beg

        time_pad_beg = 0
        time_pad_end = 0

        return add_sequential_layer(
            "%s_pad" % prefix,
            {
                "class": "pad",
                "axes": ("s:0", "s:1"),
                "padding": [(time_pad_beg, time_pad_end), (feature_pad_end, feature_pad_end)],
            },
        )

    def conv2d_fixed_padding(
        prefix, filters, kernel_size, strides, dilation_rate, data_format, conv_time_dim, source=None
    ):
        """Strided 2-D convolution with explicit padding."""
        fixed_padding("%s_pad" % prefix, kernel_size, data_format, conv_time_dim)

        padding = "VALID"
        strides = (1, strides) if conv_time_dim else strides
        filter_size = (kernel_size, kernel_size)
        dilation_rate = (dilation_rate, 1) if conv_time_dim else (1, 1)

        if data_format == "channels_first":
            NCHW = True
        else:
            NCHW = False
        add_sequential_layer(
            "%s_conv" % prefix,
            {
                "class": "conv",
                "n_out": filters,
                "filter_size": filter_size,
                "auto_use_channel_first": NCHW,
                "strides": strides,
                "dilation_rate": dilation_rate,
                "padding": padding,
                "activation": None,
                "with_bias": False,
                "dropout": 0,
                "forward_weights_init": "xavier",
                "L2": L2,
            },
            from_=source,
        )
        return "%s_conv" % prefix

    def _building_block_v2(
        prefix,
        filters,
        projection_shortcut,
        strides,
        dilation_rate,
        dilation_rate_multiplier,
        kernel_size,
        data_format,
        conv_time_dim,
    ):
        """A single block for ResNet v2, without a bottleneck.

        Batch normalization then ReLu then convolution as described by:
        Identity Mappings in Deep Residual Networks
        https://arxiv.org/pdf/1603.05027.pdf
        by Kaiming He, Xiangyu Zhang, Shaoqing Ren, and Jian Sun, Jul 2016.
        """
        add_sequential_layer("%s_in" % prefix, {"class": "copy"})
        # add_sequential_layer("%s_bn1" % prefix, {"class": "batch_norm", "masked_time": False, "fused_bn": True})
        add_sequential_layer("%s_relu1" % prefix, {"class": "activation", "activation": "relu", "batch_norm": False})
        if conv_time_dim:
            # Workaround (conv can't work with strides > 1 and dilation > 1)
            # => do striding before via slicing.
            conv2d_fixed_padding(
                prefix=("%s_conv_1" % prefix),
                filters=filters,
                kernel_size=kernel_size,
                strides=1,
                dilation_rate=dilation_rate,
                data_format=data_format,
                conv_time_dim=conv_time_dim,
            )
            add_sequential_layer("%s_stride" % prefix, {"class": "slice", "axis": "s:1", "slice_step": strides})
            dilation_rate *= dilation_rate_multiplier
        else:
            conv2d_fixed_padding(
                prefix=("%s_conv_1" % prefix),
                filters=filters,
                kernel_size=kernel_size,
                strides=strides,
                dilation_rate=dilation_rate,
                data_format=data_format,
                conv_time_dim=conv_time_dim,
            )
        # add_sequential_layer("%s_bn2" % prefix, {"class": "batch_norm", "masked_time": False, "fused_bn": True})
        add_sequential_layer("%s_relu2" % prefix, {"class": "activation", "activation": "relu", "batch_norm": False})

        conv = conv2d_fixed_padding(
            prefix=("%s_conv_2" % prefix),
            filters=filters,
            kernel_size=kernel_size,
            strides=1,
            dilation_rate=dilation_rate,
            data_format=data_format,
            conv_time_dim=conv_time_dim,
        )
        result = "%s_conv_2" % prefix
        # The projection shortcut should come after the first batch norm and ReLU
        # since it performs a 1x1 convolution.
        crop_lr = filter_size[0] - 1
        crop_left = crop_lr // 2
        crop_right = crop_lr - crop_left

        if conv_time_dim:
            if dilation_rate_multiplier > 1:
                crop = int(crop_left * (dilation_rate / dilation_rate_multiplier + dilation_rate))
            else:
                crop = int(crop_left * 2 * dilation_rate)
            add_sequential_layer(
                "%s_crop" % prefix,
                {"class": "slice", "axis": "T", "slice_start": crop, "slice_end": -crop},
                from_=("%s_relu1" % prefix),
            )
            shortcut = "%s_crop" % prefix
            if projection_shortcut is not None:
                shortcut = projection_shortcut(source=shortcut)
        else:
            crop = crop_left
            add_sequential_layer(
                "%s_crop_1" % prefix,
                {"class": "slice", "axis": "T", "slice_start": crop, "slice_end": -crop},
                from_=("%s_relu1" % prefix),
            )
            shortcut = "%s_crop_1" % prefix

            if projection_shortcut is not None:
                shortcut = projection_shortcut(source=shortcut)

            add_sequential_layer(
                "%s_crop_2" % prefix,
                {"class": "slice", "axis": "T", "slice_start": crop, "slice_end": -crop},
                from_=shortcut,
            )
            shortcut = "%s_crop_2" % prefix

        add_sequential_layer("%s_out" % prefix, {"class": "combine", "kind": "add"}, from_=[conv, shortcut])
        return

    def block_layer(
        prefix,
        filters,
        bottleneck,
        block_fn,
        blocks,
        strides,
        dilation_rate,
        dilation_rate_multiplier,
        kernel_size,
        data_format,
        conv_time_dim,
    ):
        """Creates one layer of blocks for the ResNet model."""
        # Bottleneck blocks end with 4x the number of filters as they start with
        filters_out = filters * 4 if bottleneck else filters
        if not conv_time_dim:
            strides = (dilation_rate_multiplier, strides)

        def projection_shortcut(source=None):
            return conv2d_fixed_padding(
                prefix=("%s_sc" % prefix),
                filters=filters_out,
                kernel_size=1,
                strides=strides,
                dilation_rate=1,
                data_format=data_format,
                conv_time_dim=conv_time_dim,
                source=source,
            )

        # Only the first block per block_layer uses projection_shortcut and strides
        block_fn(
            "%s_0" % prefix,
            filters,
            projection_shortcut,
            strides,
            dilation_rate,
            dilation_rate_multiplier,
            kernel_size,
            data_format,
            conv_time_dim,
        )
        dilation_rate *= dilation_rate_multiplier
        for i in range(1, blocks):
            block_fn("%s_%i" % (prefix, i), filters, None, 1, dilation_rate, 1, kernel_size, data_format, conv_time_dim)

        return add_sequential_layer("%s_out" % prefix, {"class": "copy"})

    # Params for
    # ResNet Version (d): https://arxiv.org/pdf/1703.02136.pdf
    resnet_version = 2
    conv_time_dim = conv_time_dim
    bottleneck = False
    num_filters = 64
    first_kernel_size = 5
    kernel_size = 3
    conv_stride = 2 if conv_time_dim else (1, 2)
    first_pool_size = (1, 2)
    first_pool_stride = (1, 1)
    last_pool_size = (2, 2)
    last_pool_stride = (1, 2) if conv_time_dim else (2, 2)
    block_sizes = [2, 2, 2, 2]
    block_strides = [1, 2, 2, 2]
    block_dilations = [1, 1, 1, 2]
    block_fn = _building_block_v2
    data_format = "channels_first"
    pre_activation = resnet_version == 2

    if data_format == "channels_first":
        NCHW = True
    else:
        NCHW = False
    # Calculating the reduction of the time dim
    if conv_time_dim:
        multiplier = 1 if bottleneck else 2
        building_block_reduction = multiplier * 2 * (kernel_size // 2)
        total_reduction = first_kernel_size - 1

        dilation_rate_multiplier = 1
        total_reduction += dilation_rate_multiplier * (first_pool_size[0] - 1)

        for i, bs in enumerate(block_sizes):
            total_reduction += building_block_reduction / multiplier * dilation_rate_multiplier
            dilation_rate_multiplier *= block_dilations[i]
            total_reduction += building_block_reduction / multiplier * dilation_rate_multiplier
            total_reduction += building_block_reduction * (bs - 1) * dilation_rate_multiplier

        total_reduction += dilation_rate_multiplier * (last_pool_size[0] - 1)
        dilation_rate_multiplier *= 2
        print(total_reduction, dilation_rate_multiplier)

        total_reduction += dilation_rate_multiplier * 2
        print(total_reduction, dilation_rate_multiplier)

        time_dim_reduction = total_reduction
        context_window = int(2 * (total_reduction // 2) + 1)
    else:
        time_dim_reduction = 0

    print("time_dim_reduction: ", time_dim_reduction)
    print("context_window: ", context_window)

    # Building the ResNet
    conv2d_fixed_padding(
        prefix="c_init",
        filters=num_filters,
        kernel_size=first_kernel_size,
        strides=conv_stride,
        dilation_rate=1,
        data_format=data_format,
        conv_time_dim=conv_time_dim,
    )

    dilation_rate = 1

    if resnet_version == 1:
        # add_sequential_layer("c_init_bn", {"class": "batch_norm", "masked_time": False, "fused_bn": True})
        add_sequential_layer("c_init_relu", {"class": "activation", "activation": "relu", "batch_norm": False})

    if first_pool_size:
        if conv_time_dim:
            dr = (dilation_rate, 1)
        else:
            dr = (1, 1)

        pad_total = first_pool_size[1] - 1
        pad_beg = pad_total // 2
        pad_end = pad_total - pad_beg

        add_sequential_layer("c_init_pool_pad", {"class": "pad", "axes": "s:1", "padding": (pad_beg, pad_end)})
        add_sequential_layer(
            "c_init_pool",
            {
                "class": "pool",
                "mode": "max",
                "padding": "VALID",
                "pool_size": first_pool_size,
                "strides": first_pool_stride,
                "dilation_rate": dr,
                "use_channel_first": NCHW,
            },
        )
    print("dr: ", dilation_rate)
    for i, num_blocks in enumerate(block_sizes):
        filters = num_filters * (2**i)
        block_layer(
            prefix="c_%i" % i,
            filters=filters,
            bottleneck=bottleneck,
            block_fn=block_fn,
            blocks=num_blocks,
            strides=block_strides[i],
            dilation_rate=dilation_rate,
            dilation_rate_multiplier=block_dilations[i],
            kernel_size=kernel_size,
            data_format=data_format,
            conv_time_dim=conv_time_dim,
        )
        dilation_rate *= block_dilations[i]
        print("dr: ", dilation_rate)

    if pre_activation:
        # add_sequential_layer("c_out_bn", {"class": "batch_norm", "masked_time": False, "fused_bn": True})
        add_sequential_layer("c_out_relu", {"class": "activation", "activation": "relu", "batch_norm": False})

    if last_pool_size:
        pad_total = last_pool_size[1] - 1
        pad_beg = pad_total // 2
        pad_end = pad_total - pad_beg
        add_sequential_layer("c_last_pool_pad", {"class": "pad", "axes": "s:1", "padding": (pad_beg, pad_end)})

        if conv_time_dim:
            dr = (dilation_rate, 1)
            add_sequential_layer(
                "c_last_pool",
                {
                    "class": "pool",
                    "mode": "max",
                    "padding": "VALID",
                    "pool_size": last_pool_size,
                    "strides": (1, 1),
                    "dilation_rate": dr,
                    "use_channel_first": NCHW,
                },
            )
            add_sequential_layer("c_last_stride", {"class": "slice", "axis": "s:1", "slice_step": last_pool_stride[1]})
            dilation_rate *= 2
        else:
            dr = (1, 1)
            add_sequential_layer(
                "c_last_pool",
                {
                    "class": "pool",
                    "mode": "max",
                    "padding": "VALID",
                    "pool_size": last_pool_size,
                    "strides": last_pool_stride,
                    "dilation_rate": dr,
                    "use_channel_first": NCHW,
                },
            )
    if conv_time_dim:
        dr = (dilation_rate, 1)
    else:
        dr = (1, 1)

    """
  See https://arxiv.org/pdf/1611.09288.pdf
  Fully connected layers are equivalent to, and can be trivially replaced by,
  convolutional layers with kernel (1×1) (except the first convolution which
  has kernel size matching the output of the conv stack before being flattened
  for the fully connected layers).
  """
    add_sequential_layer(
        "fc1",
        {
            "class": "conv",
            "n_out": 2048,
            "filter_size": (3, 2),
            "auto_use_channel_first": NCHW,
            "strides": (1, 1),
            "dilation_rate": dr,
            "padding": "VALID",
            "activation": None,
            "with_bias": False,
            "dropout": 0,
            "forward_weights_init": "xavier",
            "L2": L2,
        },
    )
    add_sequential_layer(
        "fc2",
        {
            "class": "conv",
            "n_out": 2048,
            "filter_size": (1, 1),
            "auto_use_channel_first": NCHW,
            "strides": (1, 1),
            "dilation_rate": (1, 1),
            "padding": "VALID",
            "activation": None,
            "with_bias": False,
            "dropout": 0,
            "forward_weights_init": "xavier",
            "L2": L2,
        },
    )
    add_sequential_layer(
        "fc3",
        {
            "class": "conv",
            "n_out": 2048,
            "filter_size": (1, 1),
            "auto_use_channel_first": NCHW,
            "strides": (1, 1),
            "dilation_rate": (1, 1),
            "padding": "VALID",
            "activation": None,
            "with_bias": False,
            "dropout": 0,
            "forward_weights_init": "xavier",
            "L2": L2,
        },
    )
    add_sequential_layer(
        "fc4",
        {
            "class": "conv",
            "n_out": 1024,
            "filter_size": (1, 1),
            "auto_use_channel_first": NCHW,
            "strides": (1, 1),
            "dilation_rate": (1, 1),
            "padding": "VALID",
            "activation": None,
            "with_bias": False,
            "dropout": 0,
            "forward_weights_init": "xavier",
            "L2": L2,
        },
    )
    add_sequential_layer(
        "fc5",
        {
            "class": "conv",
            "n_out": num_outputs,
            "filter_size": (1, 1),
            "auto_use_channel_first": NCHW,
            "strides": (1, 1),
            "dilation_rate": (1, 1),
            "padding": "VALID",
            "activation": None,
            "with_bias": False,
            "dropout": 0,
            "forward_weights_init": "xavier",
            "L2": L2,
        },
    )

    add_sequential_layer("merge", {"class": "merge_dims", "axes": ("s:0", "s:1")})
    add_sequential_layer("swap", {"class": "swap_axes", "axis1": "s:0", "axis2": "f"})
    add_sequential_layer("output", {"class": "activation", "activation": "softmax", "loss": "ce"})
    return network, context_window


def test_ResNet():
    """Test to compare Resnet convolving (window x frequency) vs (time x frequency).
    Batch_norm layers are turned off in oder to compare, since the statistics over the
    windowed input data is a bit different from the plain input (when convolving directing
    over the time dim).
    """

    def sliding_window(seq, window_size):
        import numpy as np
        import copy

        it = iter(seq)
        win = [it.__next__() for cnt in range(window_size)]  # First window
        res_arr = []
        res_arr.append(copy.deepcopy(win))
        for e in it:  # Subsequent windows
            win[:-1] = win[1:]
            win[-1] = e
            res_arr.append(copy.deepcopy(win))
        return np.array(res_arr)

    with make_scope() as session:
        import numpy as np
        import math
        from tensorflow.python.client import timeline

        net_dict_conv_td, window_size = build_resnet(conv_time_dim=True)
        net_dict_windowed, _ = build_resnet(conv_time_dim=False)

        # Making two time-steps
        time_size = window_size + 1
        data_layer_win = Data(name="win", shape=(window_size, 64, 3), dim=3, batch_dim_axis=0, sparse=False)
        data_layer_win.placeholder = tf_compat.v1.placeholder(shape=(None, window_size, 64, 3), dtype=tf.float32)

        data_layer_nowin = Data(
            name="nowin", shape=(time_size, 64, 3), dim=3, batch_dim_axis=0, time_dim_axis=1, sparse=False
        )
        data_layer_nowin.placeholder = tf_compat.v1.placeholder(shape=(None, time_size, 64, 3), dtype=tf.float32)

        extern_data_nowin = ExternData()
        extern_data_nowin.data["data"] = data_layer_nowin
        extern_data_win = ExternData()
        extern_data_win.data["data"] = data_layer_win

        net_conv_td = TFNetwork(extern_data=extern_data_nowin)
        net_conv_td.train_flag = True

        net_conv_td.construct_from_dict(net_dict_conv_td)
        net_conv_td.initialize_params(session)

        net_windowed = TFNetwork(extern_data=extern_data_win)
        net_windowed.train_flag = True

        net_windowed.construct_from_dict(net_dict_windowed)
        net_windowed.initialize_params(session)

        data = np.random.rand(time_size, 64, 3)
        data_win = sliding_window(data, window_size)
        data = np.array([data])

        feed_dict = {data_layer_nowin.placeholder: data, data_layer_win.placeholder: data_win}

        res1, res2 = session.run(
            [net_conv_td.layers["output"].output.placeholder, net_windowed.layers["output"].output.placeholder],
            feed_dict=feed_dict,
        )

        print(res1[0][0] - res2[0][0])
        print(res1[0][1] - res2[1][0])
        assert math.isclose(np.sum(res1[0][0] - res2[0][0]), 0.0, abs_tol=1e-07)
        assert math.isclose(np.sum(res1[0][1] - res2[1][0]), 0.0, abs_tol=1e-07)


if __name__ == "__main__":
    try:
        better_exchook.install()
        if len(sys.argv) <= 1:
            for k, v in sorted(globals().items()):
                if k.startswith("test_"):
                    print("-" * 40)
                    print("Executing: %s" % k)
                    try:
                        v()
                    except unittest.SkipTest as exc:
                        print("SkipTest:", exc)
                    print("-" * 40)
            print("Finished all tests.")
        else:
            assert len(sys.argv) >= 2
            for arg in sys.argv[1:]:
                print("Executing: %s" % arg)
                if arg in globals():
                    globals()[arg]()  # assume function and execute
                else:
                    eval(arg)  # assume Python code and execute
    finally:
        import threading

        # if len(list(threading.enumerate())) > 1:
        #  print("Warning, more than one thread at exit:")
        #  better_exchook.dump_all_thread_tracebacks()



#!/usr/bin/env python3

"""
Extract state tying from dataset.
"""

from __future__ import annotations

import os
import gzip
from argparse import ArgumentParser
from pprint import pprint
from xml.etree import ElementTree
import collections
from collections import defaultdict
import typing

import _setup_returnn_env  # noqa
from returnn.datasets import init_dataset, Dataset
from returnn.datasets.lm import Lexicon, AllophoneState
from returnn.log import log
from returnn.util.basic import uniq


def get_segment_name(tree):
    """
    :param tree:
    :return:
    """

    def _m(x):
        if "name" in x.attrib:
            return x.attrib["name"]
        if x.tag == "segment":
            return "1"
        assert False, "unknown name: %r, %r" % (x, vars(x))

    return "/".join(map(_m, tree))


def iter_bliss_orth(filename):
    """
    :param str filename:
    :return:
    """
    corpus_file = open(filename, "rb")
    if filename.endswith(".gz"):
        corpus_file = gzip.GzipFile(fileobj=corpus_file)

    # noinspection PyShadowingNames
    def getelements(tag):
        """Yield *tag* elements from *filename_or_file* xml incrementally."""
        context = iter(ElementTree.iterparse(corpus_file, events=("start", "end")))
        _, root = next(context)  # get root element
        tree = [root]
        for event, elem in context:
            if event == "start":
                tree += [elem]
            elif event == "end":
                assert tree[-1] is elem
                tree = tree[:-1]
            if event == "end" and elem.tag == tag:
                yield tree, elem

    for tree, elem in getelements("segment"):
        elem_orth = elem.find("orth")
        orth_raw = elem_orth.text or ""  # should be unicode
        orth_split = orth_raw.split()
        orth = " ".join(orth_split)

        yield get_segment_name(tree + [elem]), orth


def iter_dataset_targets(dataset: Dataset):
    """
    :param dataset:
    """
    dataset.init_seq_order(epoch=1)
    seq_idx = 0
    while dataset.is_less_than_num_seqs(seq_idx):
        dataset.load_seqs(seq_idx, seq_idx + 1)
        segment_name = dataset.get_tag(seq_idx)
        targets = dataset.get_data(seq_idx, "classes")
        assert targets.ndim == 1  # sparse
        targets = targets.astype("int32")
        yield segment_name, targets
        seq_idx += 1


class OrthHandler:
    """
    Orthography handler.
    """

    allo_add_all = False  # only via lexicon

    def __init__(self, lexicon, si_label=None, allo_num_states=3, allo_context_len=1, allow_ci_in_words=True):
        """
        :param Lexicon lexicon:
        :param int si_label:
        :param int allo_num_states:
        :param int allo_context_len:
        :param bool allow_ci_in_words:
        """
        self.lexicon = lexicon
        self.phonemes = sorted(self.lexicon.phonemes.keys(), key=lambda s: self.lexicon.phonemes[s]["index"])
        self.word_boundary_phones = {-1: set(), 1: set()}
        self.phon_to_possible_ctx_via_lex = {-1: {}, 1: {}}
        for lemma in self.lexicon.lemmas.values():
            for pron in lemma["phons"]:
                phons = pron["phon"].split()
                assert phons
                self.word_boundary_phones[-1].add(phons[0])
                self.word_boundary_phones[1].add(phons[-1])
                for i in range(len(phons)):
                    ps = [phons[i + j] if (0 <= (i + j) < len(phons)) else "" for j in [-1, 0, 1]]
                    self.phon_to_possible_ctx_via_lex[1].setdefault(ps[1], set()).add(ps[2])
                    self.phon_to_possible_ctx_via_lex[-1].setdefault(ps[1], set()).add(ps[0])
        for phone in self.lexicon.phoneme_list:
            if "" in self.phon_to_possible_ctx_via_lex[-1][phone]:
                self.phon_to_possible_ctx_via_lex[-1][phone].update(self.word_boundary_phones[1])
            if "" in self.phon_to_possible_ctx_via_lex[1][phone]:
                self.phon_to_possible_ctx_via_lex[1][phone].update(self.word_boundary_phones[-1])
        if allow_ci_in_words:
            for phone in self.lexicon.phoneme_list:
                self.phon_to_possible_ctx_via_lex[-1][phone].add("")
                self.phon_to_possible_ctx_via_lex[1][phone].add("")
        self.si_lemma = self.lexicon.lemmas["[SILENCE]"]
        self.si_phone = self.si_lemma["phons"][0]["phon"]  # type: str
        self.si_label = si_label
        self.allo_num_states = allo_num_states  # e.g. 3 -> 3-state HMM
        self.allo_context_len = allo_context_len  # e.g. 1 -> one left&right, i.e. triphone

    def expected_num_labels_for_monophone_state_tying(self):
        """
        Silence has 1 state, all others have allo_num_states.

        :rtype: int
        """
        num_phones = len(self.lexicon.phonemes)
        return (num_phones - 1) * self.allo_num_states + 1

    def iter_orth(self, orth):
        """
        :param str orth:
        :return: yields lemmas
        """
        symbols = list(orth.split())
        i = 0
        while i < len(symbols):
            symbol = symbols[i]
            try:
                lemma = self.lexicon.lemmas[symbol]
            except KeyError:
                if "/" in symbol:
                    symbols[i : i + 1] = symbol.split("/")
                    continue
                if "-" in symbol:
                    symbols[i : i + 1] = symbol.split("-")
                    continue
                raise
            i += 1
            yield lemma

    def _iter_possible_ctx(self, phon_id, direction):
        """
        :param str phon_id: e.g. "aa", "aw", "uh", "z", etc.
        :param int direction: 1 or -1
        :rtype: list[tuple[str]]
        """
        if self.lexicon.phonemes[phon_id]["variation"] == "none":
            return [()]
        if self.allo_add_all:
            res = [()]  # type: typing.List[typing.Tuple[str, ...]]
            res += [
                (p,) for p in sorted(self.lexicon.phonemes.keys()) if self.lexicon.phonemes[p]["variation"] == "context"
            ]
            return res
        return [((p,) if p else ()) for p in sorted(self.phon_to_possible_ctx_via_lex[direction][phon_id])]

    def num_states_for_phone(self, phon_id):
        """
        :param str phon_id:
        :return: number of allophone states for this phone
        :rtype: int
        """
        if phon_id == self.si_phone:
            return 1
        return self.allo_num_states

    def all_allophone_variations(self, phon, states=None, all_boundary_variations=False):
        """
        :param str phon:
        :param None|list[int] states: which states to yield for this phone
        :param bool all_boundary_variations:
        :return: yields AllophoneState's
        :rtype: list[AllophoneState]
        """
        if states is None:
            states = range(self.num_states_for_phone(phon))
        if all_boundary_variations:
            boundary_variations = [0, 1, 2, 3]
        else:
            boundary_variations = [0]
        for left_ctx in self._iter_possible_ctx(phon, -1):
            for right_ctx in self._iter_possible_ctx(phon, 1):
                for state in states:
                    for boundary in boundary_variations:
                        a = AllophoneState()
                        a.id = phon
                        a.context_history = left_ctx
                        a.context_future = right_ctx
                        a.state = state
                        a.boundary = boundary
                        if not all_boundary_variations:
                            if not left_ctx:
                                a.mark_initial()
                            if not right_ctx:
                                a.mark_final()
                        yield a

    # noinspection PyMethodMayBeStatic
    def _phones_to_allos(self, phones):
        for p in phones:
            a = AllophoneState()
            a.id = p
            yield a

    def _allos_set_context(self, allos):
        if self.allo_context_len == 0:
            return
        ctx = []
        for a in allos:
            if self.lexicon.phonemes[a.id]["variation"] == "context":
                a.context_history = tuple(ctx)
                ctx += [a.id]
                ctx = ctx[-self.allo_context_len :]
            else:
                ctx = []
        ctx = []
        for a in reversed(allos):
            if self.lexicon.phonemes[a.id]["variation"] == "context":
                a.context_future = tuple(reversed(ctx))
                ctx += [a.id]
                ctx = ctx[-self.allo_context_len :]
            else:
                ctx = []

    def _allos_add_states(self, allos):
        for _a in allos:
            if _a.id == self.si_phone:
                yield _a
            else:  # non-silence
                for state in range(self.allo_num_states):
                    a = AllophoneState()
                    a.id = _a.id
                    a.context_history = _a.context_history
                    a.context_future = _a.context_future
                    a.boundary = _a.boundary
                    a.state = state
                    yield a

    def orth_to_allophone_states(self, orth):
        """
        :param str orth: orthography as a str. orth.split() should give words in the lexicon
        :rtype: list[AllophoneState]
        :returns allophone state list. those will have repetitions etc
        """
        allos = []
        for lemma in self.iter_orth(orth):
            assert len(lemma["phons"]) == 1, "TODO..."
            phon = lemma["phons"][0]
            l_allos = list(self._phones_to_allos(phon["phon"].split()))
            l_allos[0].mark_initial()
            l_allos[-1].mark_final()
            allos += l_allos
        self._allos_set_context(allos)
        allos = list(self._allos_add_states(allos))
        return allos


def main():
    """
    Main entry.
    """
    arg_parser = ArgumentParser()
    arg_parser.add_argument("--action")
    arg_parser.add_argument("--print_seq", action="store_true")
    arg_parser.add_argument("--print_allos", action="store_true")
    arg_parser.add_argument("--print_targets", action="store_true")
    arg_parser.add_argument("--dataset")
    arg_parser.add_argument("--corpus")
    arg_parser.add_argument("--lexicon", help="filename")
    arg_parser.add_argument("--silence", type=int, help="index")
    arg_parser.add_argument("--context", default=1, type=int)
    arg_parser.add_argument("--hmm_states", default=3, type=int)
    arg_parser.add_argument("--state_tying_type", help="'monophone' or 'full'")
    arg_parser.add_argument("--state_tying_output", help="filename")
    arg_parser.add_argument("--allo_add_all", action="store_true")
    args = arg_parser.parse_args()

    dataset = init_dataset(args.dataset) if args.dataset else None
    corpus = dict(iter_bliss_orth(filename=args.corpus)) if args.corpus else None
    lexicon = Lexicon(filename=args.lexicon) if args.lexicon else None
    silence_label = args.silence

    if args.action == "show_corpus":
        pprint(corpus)
        return

    print("Num phones: %i" % len(lexicon.phonemes), file=log.v1)
    print("Phones: %r" % sorted(lexicon.phonemes.keys()), file=log.v1)

    orth_handler = OrthHandler(lexicon=lexicon, allo_context_len=args.context, allo_num_states=args.hmm_states)
    map_idx_to_allo = defaultdict(set)  # type: typing.Dict[int, typing.Set[AllophoneState]]
    map_allo_to_idx = {}  # type: typing.Dict[AllophoneState, int]
    if args.allo_add_all:
        orth_handler.allo_add_all = True

    print("Num HMM states: %i" % orth_handler.allo_num_states, file=log.v1)
    if args.state_tying_type == "monophone":
        print("Monophone state tying.", file=log.v1)
        num_labels = orth_handler.expected_num_labels_for_monophone_state_tying()
        all_label_idx_are_used = True
    elif args.state_tying_type == "full":
        print("Full state tying.", file=log.v1)
        phone_idxs = {
            k: i + 1 for (i, k) in enumerate(lexicon.phoneme_list)
        }  # +1 to keep 0 reserved as the term-symbol
        for phon in lexicon.phoneme_list:
            for allo in orth_handler.all_allophone_variations(phon, all_boundary_variations=True):
                allo_idx = allo.index(
                    phone_idxs=phone_idxs,
                    num_states=orth_handler.allo_num_states,
                    context_length=orth_handler.allo_context_len,
                )
                map_idx_to_allo[allo_idx].add(allo)
        num_labels = max(map_idx_to_allo.keys()) + 1
        all_label_idx_are_used = False
    else:
        raise Exception("invalid state tying type %r" % args.state_tying_type)
    print("Num labels: %i" % num_labels, file=log.v1)

    if dataset:
        count = 0
        for segment_name, targets in iter_dataset_targets(dataset):
            count += 1
            if silence_label is None or count == 1:
                likely_silence_label = collections.Counter(targets).most_common(1)[0][0]
                if silence_label is None:
                    silence_label = likely_silence_label
                if silence_label != likely_silence_label:
                    print("warning: silence %i but likely %i" % (silence_label, likely_silence_label), file=log.v2)
                print("Silence label: %i" % silence_label, file=log.v1)
                orth_handler.si_label = silence_label
                # Monophone state tying:
                for allo in orth_handler.all_allophone_variations(orth_handler.si_phone):
                    map_idx_to_allo[silence_label].add(allo)
                    map_allo_to_idx[allo] = silence_label
            assert segment_name in corpus
            orth = corpus[segment_name]
            allo_states = orth_handler.orth_to_allophone_states(orth=orth)
            if args.print_seq:
                print("%r %r" % (segment_name, orth))
            if args.print_allos:
                print("  allophone state seq: %r" % allo_states)
            tgt_seq = [t for t in uniq(targets) if t != silence_label]
            if args.print_targets:
                print("  target seq: %r" % (tgt_seq,))
            assert len(allo_states) == len(tgt_seq), "check --hmm_states or so"
            for allo, t in zip(allo_states, tgt_seq):
                allo.boundary = 0  # do not differ between boundaries
                allos = map_idx_to_allo[t]
                if allo in map_allo_to_idx:
                    assert allo in allos, "bad mapping"
                else:
                    assert allo not in allos
                    allos.add(allo)
                    map_allo_to_idx[allo] = t
            if len(map_idx_to_allo) >= num_labels:
                assert len(map_idx_to_allo) == num_labels
                assert 0 in map_idx_to_allo
                assert num_labels - 1 in map_idx_to_allo
                print("Finished with uniq mapping after %i sequences." % count, file=log.v1)
                break
            if count % 100 == 0:
                print("Have indices: %i (num labels: %i)" % (len(map_idx_to_allo), num_labels), file=log.v1)

        print("Finished. Have indices: %i (num labels: %i)" % (len(map_idx_to_allo), num_labels), file=log.v1)
        if len(map_idx_to_allo) < num_labels:
            found = []
            not_found = []
            for p in sorted(lexicon.phonemes.keys()):
                allo = AllophoneState(p, state=0)
                if allo in map_allo_to_idx:
                    found.append(p)
                else:
                    not_found.append(p)
            print("Phonemes found: %r" % found)
            print("Phonemes not found: %r" % not_found)

    if args.state_tying_output:
        assert not os.path.exists(args.state_tying_output)
        if all_label_idx_are_used:
            assert len(map_idx_to_allo) == num_labels
            assert 0 in map_idx_to_allo
            assert num_labels - 1 in map_idx_to_allo
        f = open(args.state_tying_output, "w")
        for i, allos in sorted(map_idx_to_allo.items()):
            for allo in allos:
                f.write("%s %i\n" % (allo.format(), i))
        f.close()
        print("Wrote state tying to %r." % args.state_tying_output, file=log.v1)

    print("The end.")


if __name__ == "__main__":
    from returnn.util import better_exchook

    better_exchook.install()
    log.initialize(verbosity=[2])
    main()



"""
test logging
"""

from __future__ import annotations

import _setup_test_env  # noqa
from subprocess import Popen, PIPE, STDOUT, CalledProcessError
import re
import os
import sys
from pprint import pprint
import unittest
from returnn.util import better_exchook


__my_dir__ = os.path.dirname(os.path.abspath(__file__))
__base_dir__ = os.path.dirname(__my_dir__)
__main_entry__ = __base_dir__ + "/rnn.py"
py = sys.executable


def build_env():
    """build env"""
    env_update = os.environ.copy()
    return env_update


def run(args, input=None):
    """run subproc"""
    args = list(args)
    print("run:", args)
    # RETURNN by default outputs on stderr, so just merge both together
    p = Popen(args, stdout=PIPE, stderr=STDOUT, stdin=PIPE, env=build_env())
    out, _ = p.communicate(input=input)
    print("Return code is %i" % p.returncode)
    print("std out/err:\n---\n%s\n---\n" % out.decode("utf8"))
    if p.returncode != 0:
        raise CalledProcessError(cmd=args, returncode=p.returncode, output=out)
    return out.decode("utf8")


def filter_out(ls):
    """
    :param list[str] ls:
    :rtype: list[str]
    """
    if not isinstance(ls, list):
        ls = list(ls)
    res = []
    i = 0
    while i < len(ls):
        s = ls[i]
        # Filter some TF warnings, like:
        # 2025-01-10 12:09:34.206001: E external/local_xla/xla/stream_executor/cuda/cuda_dnn.cc:9261] Unable to ...
        # 2025-01-10 12:09:34.212039: I tensorflow/core/platform/cpu_feature_guard.cc:182] This TensorFlow binary ...
        if re.search(": [IWE] (tensorflow/|external/)", s):  # some TF warnings
            i += 1
            continue
        elif "SyntaxWarning" in s:
            i += 2
            continue
        # RuntimeWarning|FutureWarning are warnings and they include the code-line in the next output line
        if i + 1 < len(ls) and ls[i + 1].startswith("  "):
            if re.match(".*:\\d+: RuntimeWarning: numpy.*", s) or re.match(".*:\\d+: FutureWarning: .*", s):
                i += 2
                continue
        if any(
            msg in s
            for msg in [
                "Setup TF inter and intra global thread pools",
                "Collecting TensorFlow device list",
                "CUDA_VISIBLE_DEVICES is not set",
            ]
        ):
            i += 1
            continue
        if "Local devices available to TensorFlow:" in s:
            i += 1
            while i < len(ls) and re.match("^ {2}\\d+/\\d+: name:.*", ls[i]):
                i += 1
                while i < len(ls) and re.match("^ {4}.*", ls[i]):
                    i += 1
            continue
        if s.startswith("systemMemory:") or s.startswith("maxCacheSize:"):
            i += 1
            continue
        res.append(ls[i])
        i += 1
    return res


def count_start_with(ls, s):
    """
    :param list[str] ls:
    :param str s:
    :rtype: int
    """
    c = 0
    for l in ls:
        if l.startswith(s):
            c += 1
    return c


def test_filter_out():
    s = """
/home/travis/virtualenv/python2.7.14/lib/python2.7/site-packages/scipy/special/__init__.py:640: RuntimeWarning: numpy.dtype size changed, may indicate binary incompatibility. Expected 96, got 88
  from ._ufuncs import *
/home/travis/virtualenv/python2.7.14/lib/python2.7/site-packages/h5py/_hl/group.py:22: RuntimeWarning: numpy.dtype size changed, may indicate binary incompatibility. Expected 96, got 88
  from .. import h5g, h5i, h5o, h5r, h5t, h5l, h5p
RETURNN starting up, version 20180724.141845--git-7865d01, date/time 2018-07-24-13-11-47 (UTC+0000), pid 2196, cwd /home/travis/build/rwth-i6/returnn, Python /home/travis/virtualenv/python2.7.14/bin/python
faulthandler import error. No module named faulthandler
Theano: 0.9.0 (<site-package> in /home/travis/virtualenv/python2.7.14/lib/python2.7/site-packages/theano)
Task: No-operation
elapsed: 0:00:00.0001
"""  # noqa
    ls = filter(None, s.splitlines())
    ls = filter_out(ls)
    pprint(ls)
    assert len(ls) == 5


def test_returnn_startup():
    out = run([py, __main_entry__, "-x", "nop", "++use_tensorflow", "1"])
    ls = out.splitlines()
    ls = filter_out(ls)
    if not 3 <= len(ls) <= 10:  # not fixed because might change
        print("output:\n%s\n\nNum lines: %i" % ("\n".join(ls), len(ls)))
        raise Exception("unexpected output number of lines")
    assert count_start_with(ls, "RETURNN starting up, version ") == 1
    assert count_start_with(ls, "TensorFlow: ") == 1
    assert "Task: No-operation" in ls


def test_returnn_startup_verbose():
    out = run([py, __main_entry__, "-x", "nop", "++use_tensorflow", "1", "++log_verbosity", "5"])
    ls = out.splitlines()
    ls = filter_out(ls)
    if not 3 <= len(ls) <= 15:  # not fixed because might change
        print("output:\n%s\n\nNum lines: %i" % ("\n".join(ls), len(ls)))
        raise Exception("unexpected output number of lines")
    assert count_start_with(ls, "RETURNN starting up, version ") == 1
    assert count_start_with(ls, "RETURNN command line options: ") == 1
    assert count_start_with(ls, "TensorFlow: ") == 1
    assert "Task: No-operation" in ls
    assert "Quitting" in ls


def test_simple_log():
    code = """
from __future__ import annotations
print("hello stdout 1")
from returnn.log import log
log.initialize(verbosity=[], logs=[], formatter=[])
print("hello stdout 2")
print("hello log 1", file=log.v3)
print("hello log 2", file=log.v3)
  """
    out = run([py], input=code.encode("utf8"))
    assert out.splitlines() == ["hello stdout 1", "hello stdout 2", "hello log 1", "hello log 2"]


def test_StreamIO():
    import io

    buf = io.StringIO()
    assert buf.getvalue() == ""
    print("buf: %r" % buf.getvalue())

    buf.write("hello")
    print("buf: %r" % buf.getvalue())
    assert buf.getvalue() == "hello"
    buf.truncate(0)  # should not change the position, thus the buffer is empty but position is len("hello")
    print("buf: %r" % buf.getvalue())
    assert buf.getvalue() == ""

    buf.write("hello")
    print("buf: %r" % buf.getvalue())
    assert buf.getvalue() == "\x00\x00\x00\x00\x00hello"  # zero-filled
    buf.truncate(0)
    buf.seek(0)
    print("buf: %r" % buf.getvalue())
    assert buf.getvalue() == ""

    buf.write("hello")
    print("buf: %r" % buf.getvalue())
    assert buf.getvalue() == "hello"
    buf.truncate(0)
    buf.seek(0)
    print("buf: %r" % buf.getvalue())
    assert buf.getvalue() == ""


if __name__ == "__main__":
    better_exchook.install()
    if len(sys.argv) <= 1:
        for k, v in sorted(globals().items()):
            if k.startswith("test_"):
                print("-" * 40)
                print("Executing: %s" % k)
                try:
                    v()
                except unittest.SkipTest as exc:
                    print("SkipTest:", exc)
                print("-" * 40)
        print("Finished all tests.")
    else:
        assert len(sys.argv) >= 2
        for arg in sys.argv[1:]:
            print("Executing: %s" % arg)
            if arg in globals():
                globals()[arg]()  # assume function and execute
            else:
                eval(arg)  # assume Python code and execute
    #  better_exchook.dump_all_thread_tracebacks()



"""
Utils for modules
"""

from __future__ import annotations
from typing import Collection
import torch


def convert_parameters_to_buffers(
    module: torch.nn.Module, parameter_names: Collection[str], *, deep: bool = True, persistent: bool
):
    """
    :param module:
    :param parameter_names:
    :param deep: parameter_name can contain '.' to access submodules
    :param persistent: whether the buffer is persistent. if True, the buffer will be saved to the state_dict.
        passed to module.register_buffer.
    """
    for parameter_name in parameter_names:
        convert_parameter_to_buffer(module, parameter_name, deep=deep, persistent=persistent)


def convert_parameter_to_buffer(module: torch.nn.Module, parameter_name: str, *, deep: bool = True, persistent: bool):
    """
    :param module:
    :param parameter_name:
    :param deep: parameter_name can contain '.' to access submodules
    :param persistent: whether the buffer is persistent. if True, the buffer will be saved to the state_dict.
        passed to module.register_buffer.
    """
    if "." in parameter_name:
        if not deep:
            raise ValueError("parameter_name can't contain '.' when deep is False")
        module_path, _, parameter_name = parameter_name.rpartition(".")
        module = module.get_submodule(module_path)

    parameter = getattr(module, parameter_name)
    if not isinstance(parameter, torch.nn.Parameter):
        raise ValueError(f"{parameter_name} is not a torch.nn.Parameter, got type {type(parameter).__name__}")
    delattr(module, parameter_name)
    parameter.requires_grad = False
    module.register_buffer(parameter_name, parameter, persistent=persistent)



"""
tests for returnn.torch.frontend
"""

import _setup_test_env  # noqa

import numpy
import sys
import torch
import pytest
import unittest
import typing

from returnn.util import better_exchook
from returnn.tensor import Tensor, Dim
import returnn.frontend as rf


rf.select_backend_torch()


# Tensor.__eq__ is disabled as per the following error in some TF tests:
# AssertionError: unhashable type: 'Tensor'.
# See CI https://github.com/rwth-i6/returnn/actions/runs/4406240591
def test_compare_eq():
    a_raw = torch.tensor([2.0, 2.0, 2.0])
    b_raw = torch.tensor([1.0, 2.0, 3.0])

    feature_dim = Dim(3)

    a = Tensor(name="a", raw_tensor=a_raw, dims=[feature_dim], dtype="float32")
    b = Tensor(name="b", raw_tensor=b_raw, dims=[feature_dim], dtype="float32")

    result = a == b
    result_alt1 = rf.compare(a, "==", b)
    result_alt2 = rf.compare(a, "eq", b)

    assert result.raw_tensor.tolist() == [False, True, False]
    assert result_alt1.raw_tensor.tolist() == [False, True, False]
    assert result_alt2.raw_tensor.tolist() == [False, True, False]


def test_compare_ne():
    a_raw = torch.tensor([2.0, 2.0, 2.0])
    b_raw = torch.tensor([1.0, 2.0, 3.0])

    feature_dim = Dim(3)

    a = Tensor(name="a", raw_tensor=a_raw, dims=[feature_dim], dtype="float32")
    b = Tensor(name="b", raw_tensor=b_raw, dims=[feature_dim], dtype="float32")

    result = a != b
    result_alt1 = rf.compare(a, "!=", b)
    result_alt2 = rf.compare(a, "<>", b)
    result_alt3 = rf.compare(a, "not_equal", b)

    assert result.raw_tensor.tolist() == [True, False, True]
    assert result_alt1.raw_tensor.tolist() == [True, False, True]
    assert result_alt2.raw_tensor.tolist() == [True, False, True]
    assert result_alt3.raw_tensor.tolist() == [True, False, True]


def test_compare_lt():
    a_raw = torch.tensor([2.0, 2.0, 2.0])
    b_raw = torch.tensor([1.0, 2.0, 3.0])

    feature_dim = Dim(3)

    a = Tensor(name="a", raw_tensor=a_raw, dims=[feature_dim], dtype="float32")
    b = Tensor(name="b", raw_tensor=b_raw, dims=[feature_dim], dtype="float32")

    result = a < b
    result_alt1 = rf.compare(a, "<", b)
    result_alt2 = rf.compare(a, "less", b)

    assert result.raw_tensor.tolist() == [False, False, True]
    assert result_alt1.raw_tensor.tolist() == [False, False, True]
    assert result_alt2.raw_tensor.tolist() == [False, False, True]


def test_compare_le():
    a_raw = torch.tensor([2.0, 2.0, 2.0])
    b_raw = torch.tensor([1.0, 2.0, 3.0])

    feature_dim = Dim(3)

    a = Tensor(name="a", raw_tensor=a_raw, dims=[feature_dim], dtype="float32")
    b = Tensor(name="b", raw_tensor=b_raw, dims=[feature_dim], dtype="float32")

    result = a <= b
    result_alt1 = rf.compare(a, "<=", b)
    result_alt2 = rf.compare(a, "less_equal", b)

    assert result.raw_tensor.tolist() == [False, True, True]
    assert result_alt1.raw_tensor.tolist() == [False, True, True]
    assert result_alt2.raw_tensor.tolist() == [False, True, True]


def test_compare_gt():
    a_raw = torch.tensor([2.0, 2.0, 2.0])
    b_raw = torch.tensor([1.0, 2.0, 3.0])

    feature_dim = Dim(3)

    a = Tensor(name="a", raw_tensor=a_raw, dims=[feature_dim], dtype="float32")
    b = Tensor(name="b", raw_tensor=b_raw, dims=[feature_dim], dtype="float32")

    result = a > b
    result_alt1 = rf.compare(a, ">", b)
    result_alt2 = rf.compare(a, "greater", b)

    assert result.raw_tensor.tolist() == [True, False, False]
    assert result_alt1.raw_tensor.tolist() == [True, False, False]
    assert result_alt2.raw_tensor.tolist() == [True, False, False]


def test_compare_ge():
    a_raw = torch.tensor([2.0, 2.0, 2.0])
    b_raw = torch.tensor([1.0, 2.0, 3.0])

    feature_dim = Dim(3)

    a = Tensor(name="a", raw_tensor=a_raw, dims=[feature_dim], dtype="float32")
    b = Tensor(name="b", raw_tensor=b_raw, dims=[feature_dim], dtype="float32")

    result = a >= b
    result_alt1 = rf.compare(a, ">=", b)
    result_alt2 = rf.compare(a, "greater_equal", b)

    assert result.raw_tensor.tolist() == [True, True, False]
    assert result_alt1.raw_tensor.tolist() == [True, True, False]
    assert result_alt2.raw_tensor.tolist() == [True, True, False]


def test_combine_add_int_tensors():
    a_raw = torch.tensor([2, 2, 2])
    b_raw = torch.tensor([1, 2, 3])

    feature_dim = Dim(3)

    a = Tensor(name="a", raw_tensor=a_raw, dims=[feature_dim], dtype="int64")
    b = Tensor(name="b", raw_tensor=b_raw, dims=[feature_dim], dtype="int64")

    result = a + b
    result_alt1 = rf.combine(a, "+", b)
    result_alt2 = rf.combine(a, "add", b)

    assert result.raw_tensor.tolist() == pytest.approx([3, 4, 5])
    assert result_alt1.raw_tensor.tolist() == pytest.approx([3, 4, 5])
    assert result_alt2.raw_tensor.tolist() == pytest.approx([3, 4, 5])
    assert result.dtype == result_alt1.dtype == result_alt2.dtype == "int64"


def test_combine_add_float_tensors():
    a_raw = torch.tensor([2.0, 2.0, 2.0])
    b_raw = torch.tensor([1.0, 2.0, 3.0])

    feature_dim = Dim(3)

    a = Tensor(name="a", raw_tensor=a_raw, dims=[feature_dim], dtype="float32")
    b = Tensor(name="b", raw_tensor=b_raw, dims=[feature_dim], dtype="float32")

    result = a + b
    result_alt1 = rf.combine(a, "+", b)
    result_alt2 = rf.combine(a, "add", b)

    assert result.raw_tensor.tolist() == pytest.approx([3.0, 4.0, 5.0])
    assert result_alt1.raw_tensor.tolist() == pytest.approx([3.0, 4.0, 5.0])
    assert result_alt2.raw_tensor.tolist() == pytest.approx([3.0, 4.0, 5.0])
    assert result.dtype == result_alt1.dtype == result_alt2.dtype == "float32"


def test_combine_add_number_to_tensor():
    a_raw = torch.tensor([2.0, 2.0, 2.0])
    b_raw = torch.tensor(3.0)

    feature_dim = Dim(3)

    a = Tensor(name="a", raw_tensor=a_raw, dims=[feature_dim], dtype="float32")
    b = Tensor(name="b", raw_tensor=b_raw, dims=[], dtype="float32")

    result = a + b
    result_alt1 = rf.combine(a, "+", b)
    result_alt2 = rf.combine(a, "add", b)

    assert result.raw_tensor.tolist() == pytest.approx([5.0, 5.0, 5.0])
    assert result_alt1.raw_tensor.tolist() == pytest.approx([5.0, 5.0, 5.0])
    assert result_alt2.raw_tensor.tolist() == pytest.approx([5.0, 5.0, 5.0])
    assert result.dtype == result_alt1.dtype == result_alt2.dtype == "float32"


def test_combine_sub_int_tensors():
    a_raw = torch.tensor([2, 2, 2])
    b_raw = torch.tensor([1, 2, 3])

    feature_dim = Dim(3)

    a = Tensor(name="a", raw_tensor=a_raw, dims=[feature_dim], dtype="int64")
    b = Tensor(name="b", raw_tensor=b_raw, dims=[feature_dim], dtype="int64")

    result = a - b
    result_alt1 = rf.combine(a, "-", b)
    result_alt2 = rf.combine(a, "sub", b)

    assert result.raw_tensor.tolist() == pytest.approx([1, 0, -1])
    assert result_alt1.raw_tensor.tolist() == pytest.approx([1, 0, -1])
    assert result_alt2.raw_tensor.tolist() == pytest.approx([1, 0, -1])
    assert result.dtype == result_alt1.dtype == result_alt2.dtype == "int64"


def test_combine_sub_float_tensors():
    a_raw = torch.tensor([2.0, 2.0, 2.0])
    b_raw = torch.tensor([1.0, 2.0, 3.0])

    feature_dim = Dim(3)

    a = Tensor(name="a", raw_tensor=a_raw, dims=[feature_dim], dtype="float32")
    b = Tensor(name="b", raw_tensor=b_raw, dims=[feature_dim], dtype="float32")

    result = a - b
    result_alt1 = rf.combine(a, "-", b)
    result_alt2 = rf.combine(a, "sub", b)

    assert result.raw_tensor.tolist() == pytest.approx([1.0, 0.0, -1.0])
    assert result_alt1.raw_tensor.tolist() == pytest.approx([1.0, 0.0, -1.0])
    assert result_alt2.raw_tensor.tolist() == pytest.approx([1.0, 0.0, -1.0])
    assert result.dtype == result_alt1.dtype == result_alt2.dtype == "float32"


def test_combine_mul_int_tensors():
    a_raw = torch.tensor([2, 2, 2])
    b_raw = torch.tensor([1, 2, 3])

    feature_dim = Dim(3)

    a = Tensor(name="a", raw_tensor=a_raw, dims=[feature_dim], dtype="int64")
    b = Tensor(name="b", raw_tensor=b_raw, dims=[feature_dim], dtype="int64")

    result = a * b
    result_alt1 = rf.combine(a, "*", b)
    result_alt2 = rf.combine(a, "mul", b)

    assert result.raw_tensor.tolist() == pytest.approx([2, 4, 6])
    assert result_alt1.raw_tensor.tolist() == pytest.approx([2, 4, 6])
    assert result_alt2.raw_tensor.tolist() == pytest.approx([2, 4, 6])
    assert result.dtype == result_alt1.dtype == result_alt2.dtype == "int64"


def test_combine_mul_float_tensors():
    a_raw = torch.tensor([2.0, 2.0, 2.0])
    b_raw = torch.tensor([1.0, 2.0, 3.0])

    feature_dim = Dim(3)

    a = Tensor(name="a", raw_tensor=a_raw, dims=[feature_dim], dtype="float32")
    b = Tensor(name="b", raw_tensor=b_raw, dims=[feature_dim], dtype="float32")

    result = a * b
    result_alt1 = rf.combine(a, "*", b)
    result_alt2 = rf.combine(a, "mul", b)

    assert result.raw_tensor.tolist() == pytest.approx([2.0, 4.0, 6.0])
    assert result_alt1.raw_tensor.tolist() == pytest.approx([2.0, 4.0, 6.0])
    assert result_alt2.raw_tensor.tolist() == pytest.approx([2.0, 4.0, 6.0])
    assert result.dtype == result_alt1.dtype == result_alt2.dtype == "float32"


def test_combine_truediv_float_tensors():
    a_raw = torch.tensor([2.0, 2.0, 2.0])
    b_raw = torch.tensor([1.0, 2.0, 3.0])

    feature_dim = Dim(3)

    a = Tensor(name="a", raw_tensor=a_raw, dims=[feature_dim], dtype="float32")
    b = Tensor(name="b", raw_tensor=b_raw, dims=[feature_dim], dtype="float32")

    result = a / b
    result_alt1 = rf.combine(a, "/", b)
    result_alt2 = rf.combine(a, "truediv", b)

    assert result.raw_tensor.tolist() == pytest.approx([2.0, 1.0, 2.0 / 3.0])
    assert result_alt1.raw_tensor.tolist() == pytest.approx([2.0, 1.0, 2.0 / 3.0])
    assert result_alt2.raw_tensor.tolist() == pytest.approx([2.0, 1.0, 2.0 / 3.0])
    assert result.dtype == result_alt1.dtype == result_alt2.dtype == "float32"


def test_combine_floordiv_int_tensors():
    a_raw = torch.tensor([2, 2, 2])
    b_raw = torch.tensor([1, 2, 3])

    feature_dim = Dim(3)

    a = Tensor(name="a", raw_tensor=a_raw, dims=[feature_dim], dtype="int64")
    b = Tensor(name="b", raw_tensor=b_raw, dims=[feature_dim], dtype="int64")

    result = a // b
    result_alt1 = rf.combine(a, "//", b)
    result_alt2 = rf.combine(a, "floordiv", b)

    assert result.raw_tensor.tolist() == pytest.approx([2, 1, 0])
    assert result_alt1.raw_tensor.tolist() == pytest.approx([2, 1, 0])
    assert result_alt2.raw_tensor.tolist() == pytest.approx([2, 1, 0])
    assert result.dtype == result_alt1.dtype == result_alt2.dtype == "int64"


def test_combine_floordiv_float_tensors():
    a_raw = torch.tensor([2.0, 2.0, 2.0])
    b_raw = torch.tensor([1.0, 2.0, 3.0])

    feature_dim = Dim(3)

    a = Tensor(name="a", raw_tensor=a_raw, dims=[feature_dim], dtype="float32")
    b = Tensor(name="b", raw_tensor=b_raw, dims=[feature_dim], dtype="float32")

    result = a // b
    result_alt1 = rf.combine(a, "//", b)
    result_alt2 = rf.combine(a, "floordiv", b)

    assert result.raw_tensor.tolist() == pytest.approx([2.0, 1.0, 0.0])
    assert result_alt1.raw_tensor.tolist() == pytest.approx([2.0, 1.0, 0.0])
    assert result_alt2.raw_tensor.tolist() == pytest.approx([2.0, 1.0, 0.0])
    assert result.dtype == result_alt1.dtype == result_alt2.dtype == "float32"


def test_combine_mod_int_tensors():
    a_raw = torch.tensor([2, 2, 2, 17])
    b_raw = torch.tensor([1, 2, 3, 4])

    feature_dim = Dim(4)

    a = Tensor(name="a", raw_tensor=a_raw, dims=[feature_dim], dtype="int64")
    b = Tensor(name="b", raw_tensor=b_raw, dims=[feature_dim], dtype="int64")

    result = a % b
    result_alt1 = rf.combine(a, "%", b)
    result_alt2 = rf.combine(a, "mod", b)

    assert result.raw_tensor.tolist() == pytest.approx([0, 0, 2, 1])
    assert result_alt1.raw_tensor.tolist() == pytest.approx([0, 0, 2, 1])
    assert result_alt2.raw_tensor.tolist() == pytest.approx([0, 0, 2, 1])
    assert result.dtype == result_alt1.dtype == result_alt2.dtype == "int64"


def test_combine_mod_float_tensors():
    a_raw = torch.tensor([2.0, 2.0, 2.0, 17.0])
    b_raw = torch.tensor([1.0, 2.0, 3.0, 4.0])

    feature_dim = Dim(4)

    a = Tensor(name="a", raw_tensor=a_raw, dims=[feature_dim], dtype="float32")
    b = Tensor(name="b", raw_tensor=b_raw, dims=[feature_dim], dtype="float32")

    result = a % b
    result_alt1 = rf.combine(a, "%", b)
    result_alt2 = rf.combine(a, "mod", b)

    assert result.raw_tensor.tolist() == pytest.approx([0.0, 0.0, 2.0, 1.0])
    assert result_alt1.raw_tensor.tolist() == pytest.approx([0.0, 0.0, 2.0, 1.0])
    assert result_alt2.raw_tensor.tolist() == pytest.approx([0.0, 0.0, 2.0, 1.0])
    assert result.dtype == result_alt1.dtype == result_alt2.dtype == "float32"


def test_combine_pow_int_tensors():
    a_raw = torch.tensor([2, 2, 2])
    b_raw = torch.tensor([1, 2, 3])

    feature_dim = Dim(3)

    a = Tensor(name="a", raw_tensor=a_raw, dims=[feature_dim], dtype="int64")
    b = Tensor(name="b", raw_tensor=b_raw, dims=[feature_dim], dtype="int64")

    result = a**b
    result_alt1 = rf.combine(a, "**", b)
    result_alt2 = rf.combine(a, "pow", b)

    assert result.raw_tensor.tolist() == pytest.approx([2, 4, 8])
    assert result_alt1.raw_tensor.tolist() == pytest.approx([2, 4, 8])
    assert result_alt2.raw_tensor.tolist() == pytest.approx([2, 4, 8])
    assert result.dtype == result_alt1.dtype == result_alt2.dtype == "int64"


def test_combine_pow_float_tensors():
    a_raw = torch.tensor([2.0, 2.0, 2.0])
    b_raw = torch.tensor([1.0, 2.0, 3.0])

    feature_dim = Dim(3)

    a = Tensor(name="a", raw_tensor=a_raw, dims=[feature_dim], dtype="float32")
    b = Tensor(name="b", raw_tensor=b_raw, dims=[feature_dim], dtype="float32")

    result = a**b
    result_alt1 = rf.combine(a, "**", b)
    result_alt2 = rf.combine(a, "pow", b)

    assert result.raw_tensor.tolist() == pytest.approx([2.0, 4.0, 8.0])
    assert result_alt1.raw_tensor.tolist() == pytest.approx([2.0, 4.0, 8.0])
    assert result_alt2.raw_tensor.tolist() == pytest.approx([2.0, 4.0, 8.0])
    assert result.dtype == result_alt1.dtype == result_alt2.dtype == "float32"


def test_combine_max():
    a_raw = torch.tensor([2.0, 2.0, 2.0])
    b_raw = torch.tensor([1.0, 2.0, 3.0])

    feature_dim = Dim(3)

    a = Tensor(name="a", raw_tensor=a_raw, dims=[feature_dim], dtype="float32")
    b = Tensor(name="b", raw_tensor=b_raw, dims=[feature_dim], dtype="float32")

    result = rf.combine(a, "max", b)
    result_alt1 = rf.combine(a, "maximum", b)

    assert result.raw_tensor.tolist() == [2.0, 2.0, 3.0]
    assert result_alt1.raw_tensor.tolist() == [2.0, 2.0, 3.0]


def test_combine_min():
    a_raw = torch.tensor([2.0, 2.0, 2.0])
    b_raw = torch.tensor([1.0, 2.0, 3.0])

    feature_dim = Dim(3)

    a = Tensor(name="a", raw_tensor=a_raw, dims=[feature_dim], dtype="float32")
    b = Tensor(name="b", raw_tensor=b_raw, dims=[feature_dim], dtype="float32")

    result = rf.combine(a, "min", b)
    result_alt1 = rf.combine(a, "minimum", b)

    assert result.raw_tensor.tolist() == [1.0, 2.0, 2.0]
    assert result_alt1.raw_tensor.tolist() == [1.0, 2.0, 2.0]


class _CheckNoPythonCalls:
    """
    Check that there is no Python code executed via sys.settrace.
    """

    def __init__(self):
        self.num_calls = 0
        self.old_tracefunc = None

    def _tracefunc(self, frame, event, arg):
        print("*** trace:", frame, event, arg)
        if frame.f_globals is vars(typing):
            print("   (ignore typing module)")
            return
        if frame.f_code is Tensor.__init__.__code__:
            print("   (ignoring Tensor.__init__ for now, remains to be implemented...)")  # TODO
            return
        if frame.f_code is _CheckNoPythonCalls.__exit__.__code__:
            print("   (ignoring _CheckNoPythonCalls.__exit__)")
            return
        self.num_calls += 1

    def __enter__(self):
        self.old_tracefunc = sys.gettrace()
        sys.settrace(self._tracefunc)
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        sys.settrace(self.old_tracefunc)
        assert self.num_calls == 0


def test_native_is_raw_torch_tensor_type():
    raw_tensor = torch.zeros(2, 3)
    raw_parameter = torch.nn.Parameter(torch.zeros(2, 3))
    numpy_tensor = numpy.zeros((2, 3))

    from returnn.frontend import _native

    mod = _native.get_module()

    with _CheckNoPythonCalls():
        assert mod.is_raw_torch_tensor_type(type(raw_tensor)) is True
        assert mod.is_raw_torch_tensor_type(type(raw_parameter)) is True
        assert mod.is_raw_torch_tensor_type(type(numpy_tensor)) is False
        assert mod.is_raw_torch_tensor_type(type(43)) is False
        assert mod.is_raw_torch_tensor_type(43) is False  # current behavior - might also raise exception instead


def test_native_get_out_permutation_to_dims():
    batch_dim = Dim(2, name="batch_dim")
    time_dim = Dim(3, name="time_dim")
    feature_dim = Dim(5, name="feature_dim")
    tensor_f = Tensor(name="x", dims=[feature_dim], dtype="float32")
    tensor_bf = Tensor(name="x", dims=[batch_dim, feature_dim], dtype="float32")
    tensor_bft = Tensor(name="x", dims=[batch_dim, feature_dim, time_dim], dtype="float32")

    with _CheckNoPythonCalls():
        assert tensor_f.get_out_permutation_to_dims([feature_dim]) == [0]
        assert tensor_f.get_out_permutation_to_dims([batch_dim, feature_dim, time_dim]) == [-1, 0, -1]
        assert tensor_f.get_out_permutation_to_dims([feature_dim, time_dim]) == [0, -1]
        assert tensor_f.get_out_permutation_to_dims([time_dim, feature_dim]) == [-1, 0]
        assert tensor_f.get_out_permutation_to_dims([batch_dim, time_dim, feature_dim]) == [-1, -1, 0]
        assert tensor_bf.get_out_permutation_to_dims([batch_dim, feature_dim]) == [0, 1]
        assert tensor_bf.get_out_permutation_to_dims([feature_dim, batch_dim]) == [1, 0]
        assert tensor_bf.get_out_permutation_to_dims([batch_dim, feature_dim, time_dim]) == [0, 1, -1]
        assert tensor_bf.get_out_permutation_to_dims([feature_dim, time_dim, batch_dim]) == [1, -1, 0]
        assert tensor_bft.get_out_permutation_to_dims([batch_dim, feature_dim, time_dim]) == [0, 1, 2]
        assert tensor_bft.get_out_permutation_to_dims([feature_dim, batch_dim, time_dim]) == [1, 0, 2]

    try:
        tensor_f.get_out_permutation_to_dims([batch_dim])
    except ValueError as exc:
        print("Got expected exc", exc)
    else:
        assert False, "should have failed"


def test_torch_native_setup():
    tensor = Tensor(name="x", raw_tensor=torch.tensor([1.0, 2.0, 3.0]), dims=[Dim(3)], dtype="float32")

    from returnn.frontend._backend import global_backend
    from returnn.torch.frontend import TorchBackend

    assert isinstance(global_backend, TorchBackend)

    assert global_backend.executing_eagerly()
    assert TorchBackend.executing_eagerly()
    assert global_backend.get_dtype_name_raw(tensor.raw_tensor) == "float32"
    assert TorchBackend.get_dtype_name_raw(tensor.raw_tensor) == "float32"
    assert global_backend.get_ndim_raw(tensor.raw_tensor) == 1
    assert TorchBackend.get_ndim_raw(tensor.raw_tensor) == 1


def test_native_torch_raw_backend():
    tensor = Tensor(name="a", raw_tensor=torch.tensor([1.0, 2.0, 3.0]), dims=[Dim(3)], dtype="float32")
    backend1 = tensor._raw_backend

    import returnn.frontend._backend as _backend_api

    backend2 = _backend_api.get_backend_by_raw_tensor_type(type(tensor.raw_tensor))

    from returnn.frontend import _native

    mod = _native.get_module(verbose=True)

    with _CheckNoPythonCalls():
        backend3 = mod.get_backend_for_tensor(tensor)

    assert backend1 is backend2 is backend3


def test_native_torch_raw_backend_raw_dtype():
    raw = torch.tensor([1.0, 2.0, 3.0], dtype=torch.float32)

    from returnn.frontend import _native

    mod = _native.get_module()

    with _CheckNoPythonCalls():
        dtype = mod.raw_torch_tensor_get_dtype(raw)
    assert isinstance(dtype, str) and dtype == "float32"


def test_native_torch_tensor_eq():
    batch_dim = Dim(2, name="batch_dim")
    feature_dim = Dim(3, name="feature_dim")
    tensor_bf = Tensor("tensor", dims=[batch_dim, feature_dim], dtype="float32", raw_tensor=torch.zeros(2, 3))
    tensor_f = Tensor(
        "tensor", dims=[feature_dim], dtype="float32", raw_tensor=torch.arange(-1, 2, dtype=torch.float32)
    )

    from returnn.frontend import _native

    mod = _native.get_module()
    with _CheckNoPythonCalls():
        res1 = mod.tensor_eq(tensor_bf, tensor_bf)
        res2 = mod.tensor_eq(tensor_bf, tensor_f)
        res3 = mod.tensor_eq(tensor_bf, 0.0)

    assert isinstance(res1, Tensor) and isinstance(res1.raw_tensor, torch.Tensor)
    assert res1.dims == (batch_dim, feature_dim)
    assert res1.raw_tensor.detach().numpy().tolist() == [[True, True, True], [True, True, True]]
    assert isinstance(res2, Tensor) and isinstance(res2.raw_tensor, torch.Tensor)
    assert res2.dims == (batch_dim, feature_dim)
    assert res2.raw_tensor.detach().numpy().tolist() == [[False, True, False], [False, True, False]]
    assert isinstance(res3, Tensor) and isinstance(res3.raw_tensor, torch.Tensor)
    assert res3.dims == (batch_dim, feature_dim)
    assert res3.raw_tensor.detach().numpy().tolist() == [[True, True, True], [True, True, True]]


def test_native_torch_tensor_eq_op():
    batch_dim = Dim(2, name="batch_dim")
    feature_dim = Dim(3, name="feature_dim")
    tensor_bf = Tensor("tensor", dims=[batch_dim, feature_dim], dtype="float32", raw_tensor=torch.zeros(2, 3))
    tensor_f = Tensor(
        "tensor", dims=[feature_dim], dtype="float32", raw_tensor=torch.arange(-1, 2, dtype=torch.float32)
    )

    from returnn.frontend import _native

    mod = _native.get_module()

    assert Tensor.__eq__ is mod.tensor_eq

    with _CheckNoPythonCalls():
        res = tensor_bf == tensor_f

    assert isinstance(res, Tensor) and isinstance(res.raw_tensor, torch.Tensor)
    assert res.dims == (batch_dim, feature_dim)
    assert res.raw_tensor.detach().numpy().tolist() == [[False, True, False], [False, True, False]]


def test_native_torch_tensor_neg():
    batch_dim = Dim(2, name="batch_dim")
    feature_dim = Dim(3, name="feature_dim")
    tensor = Tensor("tensor", dims=[batch_dim, feature_dim], dtype="float32", raw_tensor=torch.ones(2, 3))

    from returnn.frontend import _native

    mod = _native.get_module()

    with _CheckNoPythonCalls():
        res = mod.tensor_neg(tensor)

    assert isinstance(res, Tensor) and isinstance(res.raw_tensor, torch.Tensor)
    assert res.dims == (batch_dim, feature_dim)
    assert res.raw_tensor.detach().numpy().tolist() == [[-1.0, -1.0, -1.0], [-1.0, -1.0, -1.0]]


def test_native_torch_tensor_sub():
    batch_dim = Dim(2, name="batch_dim")
    feature_dim = Dim(3, name="feature_dim")
    tensor_bf = Tensor("tensor", dims=[batch_dim, feature_dim], dtype="float32", raw_tensor=torch.ones(2, 3))
    tensor_f = Tensor(
        "tensor", dims=[feature_dim], dtype="float32", raw_tensor=torch.arange(-1, 2, dtype=torch.float32)
    )

    from returnn.frontend import _native

    mod = _native.get_module()

    with _CheckNoPythonCalls():
        res1 = mod.tensor_sub(tensor_bf, tensor_bf)
        res2 = mod.tensor_sub(tensor_bf, tensor_f)
        res3 = mod.tensor_sub(tensor_bf, 3.0)

    assert isinstance(res1, Tensor) and isinstance(res1.raw_tensor, torch.Tensor)
    assert res1.dims == (batch_dim, feature_dim)
    assert res1.raw_tensor.detach().numpy().tolist() == [[0.0, 0.0, 0.0], [0.0, 0.0, 0.0]]
    assert isinstance(res2, Tensor) and isinstance(res2.raw_tensor, torch.Tensor)
    assert res2.dims == (batch_dim, feature_dim)
    assert res2.raw_tensor.detach().numpy().tolist() == [[2.0, 1.0, 0.0], [2.0, 1.0, 0.0]]
    assert isinstance(res3, Tensor) and isinstance(res3.raw_tensor, torch.Tensor)
    assert res3.dims == (batch_dim, feature_dim)
    assert res3.raw_tensor.detach().numpy().tolist() == [[-2.0, -2.0, -2.0], [-2.0, -2.0, -2.0]]


def test_native_torch_tensor_sub_permute_more_dims():
    # Make sure at least 3 dims need to be permuted,
    # such that the inverse permutation would be wrong.
    batch_dim = Dim(2, name="batch_dim")
    time_dim = Dim(3, name="time_dim")
    feature_dim = Dim(5, name="feature_dim")
    tensor_bft = Tensor(
        "tensor",
        dims=[batch_dim, feature_dim, time_dim],
        dtype="int32",
        raw_tensor=torch.arange(1, 1 + 2 * 3 * 5, dtype=torch.int32).reshape(2, 5, 3),
    )
    tensor_tbf = tensor_bft.copy_transpose([time_dim, batch_dim, feature_dim])  # permute [2,0,1]
    assert tensor_tbf.dims == (time_dim, batch_dim, feature_dim)
    # Inverse permute TBF -> BFT would be [1,2,0].

    from returnn.frontend import _native

    mod = _native.get_module()

    with _CheckNoPythonCalls():
        res1 = mod.tensor_sub(tensor_bft, tensor_tbf)

    assert isinstance(res1, Tensor) and isinstance(res1.raw_tensor, torch.Tensor)
    assert res1.dims == (batch_dim, feature_dim, time_dim)
    assert all(0 == v for v in res1.raw_tensor.detach().numpy().flatten().tolist())


# TODO test case for reshape more dims

# TODO test case for reshape + permute, make sure permute inverse would be wrong (2 permute dims not enough)

# TODO test case for generic with allow_broadcast_all_sources=True

# TODO test case for Dim with declare_same_as (so fast paths would not apply, fallback to generic)

# TODO test case with duplicated dims with match_priority


if __name__ == "__main__":
    better_exchook.install()
    if len(sys.argv) <= 1:
        for k, v in sorted(globals().items()):
            if k.startswith("test_"):
                print("-" * 40)
                print("Executing: %s" % k)
                try:
                    v()
                except unittest.SkipTest as exc:
                    print("SkipTest:", exc)
                print("-" * 40)
        print("Finished all tests.")
    else:
        assert len(sys.argv) >= 2
        for arg in sys.argv[1:]:
            print("Executing: %s" % arg)
            if arg in globals():
                globals()[arg]()  # assume function and execute
            else:
                eval(arg)  # assume Python code and execute
